from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from .terminalbase import TerminalBase


class AI:
    def __init__(self, terminal: TerminalBase) -> None:
        self.terminal = terminal
        self.base = terminal.base

        self.prompt = f"""You are an expert at terminal commands. you know all the commands
                and their arguments. You can generate correct commands with arguments from the command
                that was not recognized by the terminal. You can recognize and fix commands from any shell 
                in any operating system. Only return the fixed command, and dont explain it.

                The user may talk in english as well, its when he is not sure how to write the command.
                You can also generate commands for the user if he is not sure what to do!
                
                Shell executable used: {self.terminal.shell}
                Shell name: {self.terminal.name}
                System details: {str(self.base.system)}
                Input from user: """

        self.model = self.base.ai.get_model_instance(self.prompt)
        self.model.start_chat()

    def get_response(self, err_command: str) -> str:
        try:
            response = self.model.send_message(f"{self.prompt} {err_command}")
            if response:
                self.base.palette.show("runc:", response.strip())
        except Exception as e:
            self.base.logger.error(e)
            self.base.notifications.error(
                "Error occurred while fetching response. Please try again.",
                actions=[
                    (
                        "Configuration",
                        lambda: self.base.sidebar.show_ai().add_placeholder(),
                    )
                ],
            )



from __future__ import annotations

import os
import subprocess
import sys
import threading
import typing
from tkinter import filedialog
from tkinter.messagebox import askyesnocancel

from biscuit.common.actionset import ActionSet

from .common import BaseGame
from .config import ConfigManager
from .editor import *
from .gui import GUIManager
from .layout import *
from .settings import *
from .views import *

if typing.TYPE_CHECKING:
    from biscuit.editor import BaseEditor, Editor
    from biscuit.language.data import *

    from .layout import *


class EventManager(GUIManager, ConfigManager):
    """
    EVENT MANAGER
    -------------

    Event manager part of Biscuit Core.
    Manages the application events, actions, and interactions.
    """

    menubar: Menubar
    statusbar: Statusbar

    contentpane: Content
    editorsmanager: EditorsManager

    sidebar: SideBar
    secondary_sidebar: SecondarySideBar
    explorer: Explorer
    search: Search
    outline: Outline
    source_control: SourceControl
    extensionsGUI: Extensions

    panel: Panel
    terminalmanager: Terminal
    logger: Logs

    def set_title(self, title: str = None) -> None:
        if not self.initialized:
            return
        self.menubar.change_title(title)

    def open(self, path: str, warn_for_directory=False) -> None:
        """Opens file/directory based on path.

        TODO: Open directory in new window if warn_for_directory is True.
        """
        if not path:
            return

        if os.path.isdir(path):
            return self.open_directory(path)

        if os.path.isfile(path):
            return self.open_editor(path)

        self.notifications.error(f"Path does not exist: {path}")

    def open_directory(self, dir: str) -> None:
        if not dir or not os.path.isdir(dir):
            return

        self.statusbar.process_indicator.show()
        self.active_directory = dir

        self.explorer.directory.change_path(dir, create_root=False)
        self.set_title(os.path.basename(self.active_directory))

        self.editorsmanager.delete_all_editors()
        self.terminalmanager.delete_all_terminals()
        self.terminalmanager.open_terminal()

        try:
            self.git.check_git()
            self.update_git()
        except Exception as e:
            self.logger.error(f"Checking git failed: {e}")
            self.notifications.error("Checking git failed: see logs")

        self.explorer.directory.create_root(self.active_directory, subdir=False)

        self.event_generate("<<DirectoryChanged>>", data=dir)

    def update_git(self) -> None:
        self.statusbar.update_git_info()
        self.source_control.refresh()

    def clone_repo(self, url: str, new_window: bool = True) -> None:
        path = filedialog.askdirectory()
        if not path:
            return

        if new_window:
            new_window = askyesnocancel(
                "Open in new window or current",
                "Do you want to open the cloned repository in a new window?",
            )

        try:

            def clone() -> None:
                repodir = self.git.clone(url, path)
                if new_window:
                    self.open_in_new_window(repodir)
                else:
                    self.open_directory(repodir)

            temp = threading.Thread(target=clone)
            temp.daemon = True
            temp.start()

        except Exception as e:
            self.logger.error(f"Cloning repository failed: {e}")
            self.notifications.error("Cloning repository failed: see logs")
            return

    def close_active_directory(self) -> None:
        self.active_directory = None
        self.explorer.directory.close_directory()
        self.editorsmanager.delete_all_editors()
        self.set_title()
        self.git_found = False
        self.update_git()
        self.event_generate("<<DirectoryChanged>>")

    def close_active_editor(self) -> None:
        self.editorsmanager.close_active_editor()

    def goto_location_in_active_editor(self, position: str) -> None:
        if editor := self.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                editor.content.text.focus_set()
                editor.content.goto(position)

    def goto_location(self, path: str, position: str) -> None:
        if self.editorsmanager.is_open(path):
            if editor := self.editorsmanager.set_active_editor_by_path(path):
                editor.content.text.focus_set()
                editor.content.goto(position)
            return

        editor = self.open_editor(path, exists=True, load_file=False)
        editor.content.bind(
            "<<FileLoaded>>", lambda e, pos=position: editor.content.goto(pos), add=True
        )
        editor.content.load_file()

    def open_workspace_edit(self, path: str, edits: list[TextEdit]):
        if self.editorsmanager.is_open(path):
            e = self.editorsmanager.set_active_editor_by_path(path).content.text
            self.do_workspace_edits(e, edits)
            return

        editor = self.open_editor(path, exists=True, load_file=False)
        editor.content.bind(
            "<<FileLoaded>>",
            lambda _, editor=editor.content.text, edits=edits: threading.Thread(
                target=self.do_workspace_edits, args=(editor, edits), daemon=True
            ).start(),
            add=True,
        )
        editor.content.load_file()

    def do_workspace_edits(self, tab: Text, edits: list[TextEdit]):
        for i in edits:
            tab.replace(i.start, i.end, i.new_text)
            tab.update()
            tab.update_idletasks()

    def open_editor(
        self, path: str, exists=True, load_file=True
    ) -> Editor | BaseEditor:
        """Opens suitable editor based on the path, exists and load_file flags.

        `exists` flag will prioritize over `load_file`. If `exists` is False, `load_file` will be False.
        """

        if not path:
            exists = False

        elif not os.path.isfile(path):
            exists = False

        return self.editorsmanager.open_editor(path, exists, load_file)

    def open_files(self, paths: list[str]) -> None:
        for path in paths:
            self.open_editor(path)

    def open_diff(self, path: str, kind: str) -> None:
        # TODO kind kwarg
        self.editorsmanager.open_diff_editor(path, kind)  # type: ignore

    def diff_files(self, file1: str, file2: str) -> None:
        self.editorsmanager.diff_files(file1, file2, standalone=True)

    def open_settings(self, *_) -> None:
        self.editorsmanager.add_editor(SettingsEditor(self.editorsmanager))

    def open_game(self, name: str) -> None:
        self.editorsmanager.open_game(name)

    def register_game(self, game: BaseGame) -> None:
        # TODO game manager class
        self.game_manager.register_new_game(game)
        self.settings.generate_actionset()

    def register_langserver(self, language: str, command: str) -> None:
        self.language_server_manager.register_langserver(language, command)

    def register_comment_prefix(self, language: str, prefix: str) -> None:
        register_comment_prefix(language, prefix)

    def register_actionset(self, actionset: ActionSet) -> None:
        self.palette.register_actionset(actionset)

    def register_command(self, name: str, command: typing.Callable) -> None:
        self.settings.register_command(name, command)

    def register_run_command(self, language: str, command: str) -> None:
        self.execution_manager.register_command(language, command)

    def open_in_new_window(self, dir: str) -> None:
        subprocess.Popen([sys.executable, sys.argv[0], dir])

    def open_new_window(self) -> None:
        subprocess.Popen([sys.executable, sys.argv[0]])

        # from .main import get_app_instance
        # app = get_app_instance()
        # multiprocessing.freeze_support()
        # multiprocessing.Process(target=app.run).start()

    def workspace_opened(self) -> None:
        workspace = self.active_workspace
        try:
            self.open_directory(workspace.dirs[0])
            for dir in workspace.dirs[1:]:
                self.open_in_new_window(dir)
        except Exception as e:
            self.logger.error(f"Opening workspace failed: {e}")
            self.notifications.error("Opening workspace failed: see logs")

    def workspace_changed(self, dir: str) -> None:
        self.open_in_new_window(dir)

    def workspace_closed(self) -> None:
        self.close_active_directory()
        # TODO add cli args to flag new windows opened by workspace and close them

    def toggle_terminal(self) -> None:
        self.panel.switch_to_terminal()
        self.contentpane.toggle_panel()

    def update_statusbar(self) -> None:
        if editor := self.editorsmanager.active_editor:
            if editor.content and editor.content.editable:
                self.statusbar.toggle_editmode(True)
                active_text = editor.content.text
                self.statusbar.set_encoding(active_text.encoding)
                return self.statusbar.set_line_col_info(
                    active_text.line, active_text.column, len(active_text.selection)
                )

        self.statusbar.toggle_editmode(False)

    def log_error(self, message: str) -> None:
        self.logger.error(message)
        self.notifications.error(
            message, actions=[("Show logs", self.commands.show_logs)]
        )

    def log_info(self, message: str) -> None:
        self.logger.info(message)
        self.notifications.info(
            message, actions=[("Show logs", self.commands.show_logs)]
        )

    def log_warning(self, message: str) -> None:
        self.logger.warning(message)
        self.notifications.warning(
            message, actions=[("Show logs", self.commands.show_logs)]
        )

    def log_trace(self, message: str) -> None:
        self.logger.trace(message)



import os
import tkinter as tk
from pathlib import Path

from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, IconLabelButton

from .history_navigation import HistoryNavigation


class Item(IconLabelButton):
    """Item in the breadcrumbs widget

    Item is a button that represents a directory in the breadcrumbs widget.
    PathView is shown when the user clicks on the item."""

    def __init__(self, master, path: str, text: str, callback, *args, **kwargs) -> None:
        super().__init__(
            master,
            text=text,
            pady=10,
            icon=Icons.CHEVRON_RIGHT,
            callback=callback,
            iconside=tk.RIGHT,
            hfg_only=True,
        )
        self.path = path
        self.text_label.config(padx=0, pady=0, font=self.base.settings.uifont)
        self.icon_label.config(padx=0, pady=0, font=("codicon", 12))

    def on_click(self, e: tk.Event):
        self.callback(self)


# TODO show current symbols when language features are enabled
class BreadCrumbs(Frame):
    """Breadcrumbs widget

    Breadcrumbs widget is used to display the path of the current file.
    It is used in the editor to show the directory structure of the file.
    User can click on the breadcrumbs to navigate the directory tree."""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.editors.breadcrumbs)

        self.pathview = self.base.pathview

        self.container = Frame(self)
        self.container.pack(side=tk.LEFT, fill=tk.Y, expand=True, anchor=tk.W, padx=10)

        self.tab_control = HistoryNavigation(self)
        self.tab_control.pack(side=tk.RIGHT, fill=tk.Y)

    def set_path(self, path: str):
        active = self.base.active_directory
        try:
            # if the file belongs to active directory, use relative path instead
            if active and os.path.commonpath(
                [active, os.path.abspath(path)]
            ) == os.path.abspath(active):
                return self.add_relative(path)

        except ValueError:
            # mostly happens when paths don't have the same drive
            pass
        except Exception as e:
            print(f"Error in setting breadcrumbs: {e}")

        # otherwise, use the absolute path
        self.add_absolute(path)

    def add_absolute(self, path):
        path = Path(path).resolve()
        for i, item in enumerate(path.parts):
            text = item if item == path.parts[-1] else f"{item}"
            if not i:
                self.additem(Path(item), text)
                continue
            self.additem(Path(*path.parts[:i]), text)
        return path

    def add_relative(self, path):
        active = self.base.active_directory
        path = os.path.relpath(path, active).split(os.path.sep)
        for i, item in enumerate(path):
            text = item if item == path[-1] else f"{item}"
            self.additem(os.path.join(active, *path[:i]), text)

    def additem(self, path, text):
        btn = Item(self.container, path, text, self.pathview.show)
        btn.pack(side=tk.LEFT)

    def clear(self):
        for i in self.container.winfo_children():
            i.destroy()

    def hide(self):
        self.clear()
        self.pack_forget()

    def show(self):
        self.clear()
        self.pack(fill=tk.X, side=tk.TOP)



import tkinter as tk

from biscuit.common.icons import Icons

from ..sidebar_view import SideBarView
from .git import Git
from .menu import SourceControlMenu


class SourceControl(SideBarView):
    """The Source Control view.

    The Source Control view allows the user to manage the source control of the active document.
    Git is the only source control system supported.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.__icon__ = Icons.GIT_COMMIT
        self.name = "Source Control"

        self.git = Git(self)
        self.add_item(self.git)
        self.bind("<Visibility>", self.reload_tree)

        self.menu = SourceControlMenu(self, "files")
        self.menu.add_checkable("Show Staged", self.git.toggle_staged, checked=True)
        self.menu.add_separator(10)
        self.menu.add_checkable("Show Changes", self.git.toggle_changes, checked=True)
        self.add_action(Icons.REFRESH, self.refresh)
        self.add_action(Icons.REPO_PUSH, self.git.push)
        self.add_action(Icons.REPO_PULL, self.git.pull)
        self.add_action(Icons.ELLIPSIS, self.menu.show)

    def reload_tree(self, *_) -> None:
        self.git.open_repo()

    def refresh(self, *_) -> None:
        if self.base.git_found:
            self.git.enable_tree()
            self.git.open_repo()
        else:
            self.git.disable_tree()



import sv_ttk
from pygments.token import Token

from .theme import Theme


class VSCodeLight(Theme):
    name = "VSCode light"

    def __init__(self, *args, **kwds) -> None:
        super().__init__(*args, **kwds)
        sv_ttk.use_light_theme()



import platform
import sqlite3

from tkinterDnD import Tk

from .common import *
from .config import ConfigManager
from .editor import *
from .extensions import *
from .layout import *


class GUIManager(Tk, ConfigManager):
    """GUI manager of Biscuit core (Tkinter)
    -------------------------------------

    Initializes the underlying tkinter GUI and sets up the main components.
    - This is the root of the GUI, and all major components are initialized here.
    - DPI scaling is handled here.
    - Also gives quick access to major components of the editor from API.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize_tk(self) -> None:
        """Initialize and configure tkinter
        Handles DPI scaling and window decorations."""

        # remove decorations on windows only
        # config for high DPI with winAPI
        if platform.system() == "Windows":
            from ctypes import windll

            # TODO: DPI awareness for windows
            # disabled because it messes up the scaling
            # windll.shcore.SetProcessDpiAwareness(1)

            self.overrideredirect(True)
            self.update_idletasks()

            GWL_EXSTYLE = -20
            WS_EX_APPWINDOW = 0x00040000
            WS_EX_TOOLWINDOW = 0x00000080
            hwnd = windll.user32.GetParent(self.winfo_id())
            style = windll.user32.GetWindowLongW(hwnd, GWL_EXSTYLE)
            style = style & ~WS_EX_TOOLWINDOW
            style = style | WS_EX_APPWINDOW
            windll.user32.SetWindowLongW(hwnd, GWL_EXSTYLE, style)

            myappid = "com.tomlin7.biscuit"
            windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

        # Works in linux, but currently causing problems in windows
        # due to `shcore.SetProcessDpiAwareness(1)` so currently disabled
        self.dpi_value = self.winfo_fpixels("1i")
        self.scale = self.dpi_value / 95.90560471976401
        # self.tk.call('tk', 'scaling', self.scale)

        # window size and position
        self.min_width = round(500)
        self.min_height = round(500)
        app_width = round(1280 * self.scale)
        app_height = round(720 * self.scale)
        x = int((self.winfo_screenwidth() - app_width) / 2)
        y = int((self.winfo_screenheight() - app_height) / 2)
        self.geometry(f"{app_width}x{app_height}+{x}+{y}")

        # very bad hack to fix the window size on windows
        self.withdraw()
        self.update_idletasks()
        self.after(10, self.wm_deiconify)

        self.setup_floating_widgets()
        self.initialize_layout()

        self.protocol("WM_DELETE_WINDOW", self.on_close_app)

    def initialize_layout(self):
        """Initialize the layout of the GUI.

        - Initializes the root of the GUI.
        - Saves references to major components for easy access."""

        # the very parent of all GUI parts
        self.root = Root(self)
        self.editortypes = EditorTypes(self)

        # major components of layout
        self.menubar = self.root.menubar
        self.statusbar = self.root.statusbar

        self.sidebar = self.root.sidebar
        self.secondary_sidebar = self.root.secondary_sidebar
        self.contentpane = self.root.content

        self.panel = self.contentpane.panel
        self.editorsmanager = self.root.content.editorspane
        self.editorsbar = self.editorsmanager.editorsbar
        self.breadcrumbs = self.editorsbar.breadcrumbs

        self.explorer = self.sidebar.explorer
        self.search = self.sidebar.search
        self.outline = self.secondary_sidebar.outline
        self.source_control = self.secondary_sidebar.source_control
        self.debug = self.sidebar.debug
        self.ai = self.secondary_sidebar.ai
        self.github = self.secondary_sidebar.github
        self.extensions_view = self.secondary_sidebar.extensions

        self.terminalmanager = self.panel.terminals
        self.logger = self.panel.logger
        self.problems = self.panel.problems

    def setup_floating_widgets(self) -> None:
        """Setup floating widgets that are not part of the main layout."""

        # separate function because why not

        self.palette = Palette(self)
        self.findreplace = FindReplace(self)
        self.notifications = Notifications(self)
        self.extension_viewer = ExtensionViewer(self)

        self.text_editor_context_menu = TextEditorContextMenu(self)

        self.open_editors = OpenEditors(self)
        self.autocomplete = AutoComplete(self)
        self.peek = Peek(self)
        self.rename = Rename(self)
        self.pathview = PathView(self)
        self.hover = Hover(self)
        self.diagnostic = Diagnostic(self)

    def late_setup(self) -> None:
        """Late setup for GUI components that are dependant on other components.
        Called after initialization of major components."""

        self.binder.late_bind_all()
        self.editorsmanager.add_default_editors()
        self.editorsmanager.generate_actionsets()
        self.settings.late_setup()
        self.history.generate_actionsets()
        self.git.late_setup()
        self.debugger_manager.register_actionsets()
        self.register_misc_palettes()

        # force set focus on this window
        self.focus_set()

        if self.testing:
            return

        self.setup_extensions()

    def register_misc_palettes(self) -> None:
        """Register miscellaneous palettes that don't belong to any specific component."""

        self.google_search = ActionSet(
            "Search on google...",
            "google:",
            pinned=[
                ["Search on Google: {}", search_google],
            ],
        )
        self.palette.register_actionset(lambda: self.google_search)

    def set_title(self, title: str = None) -> None:
        """Set the title of the application window.

        Args:
            title (str, optional): Title of the window. Defaults to None."""

        if not self.initialized:
            return

        self.menubar.set_title(title)
        self.menubar.reposition_title()

    def _get_opened_files(self) -> list:
        """
        Get the file paths of all opened files from the active editors.

        Returns:
            list: A list of file paths corresponding to the opened files.
        """

        opened_files = []

        for editor in self.editorsmanager.active_editors:
            if editor.path:
                opened_files.append(editor.path)

        return opened_files

    def on_close_app(self) -> None:

        opened_files = self._get_opened_files()
        self.sessions.clear_session()
        self.sessions.save_session(opened_files, self.active_directory)

        self.editorsmanager.delete_all_editors()
        self.destroy()

    def resize(self, mode: str) -> None:
        """Resize the window based on the mode.

        Args:
            mode (str): Mode of resizing"""

        abs_x = self.winfo_pointerx() - self.winfo_rootx()
        abs_y = self.winfo_pointery() - self.winfo_rooty()
        width = self.winfo_width()
        height = self.winfo_height()
        x = self.winfo_rootx()
        y = self.winfo_rooty()

        match mode:
            case "e":
                if height > self.min_height and abs_x > self.min_width:
                    return self.geometry(f"{abs_x}x{height}")
            case "n":
                height = height - abs_y
                y = y + abs_y
                if height > self.min_height and width > self.min_width:
                    return self.geometry(f"{width}x{height}+{x}+{y}")
            case "w":
                width = width - abs_x
                x = x + abs_x
                if height > self.min_height and width > self.min_width:
                    return self.geometry(f"{width}x{height}+{x}+{y}")
            case "s":
                height = height - (height - abs_y)
                if height > self.min_height and width > self.min_width:
                    return self.geometry(f"{width}x{height}+{x}+{y}")



from __future__ import annotations

import queue
import subprocess
import typing
from threading import Thread

if typing.TYPE_CHECKING:
    from biscuit import App


class IO:
    """Handling input/output of a process in a separate thread"""

    def __init__(self, master, cmd: str, cwd: str) -> None:
        """Initialize the IO class

        Args:
            master: The parent object
            cmd (str): The command to run
            cwd (str): The working directory"""

        self.master = master
        self.base: App = master.base
        self.alive = True
        self.cmd = cmd
        self.cwd = cwd

        self.in_queue = queue.Queue()  # input data
        self.out_queue = queue.Queue()  # output results

    def write(self, buf) -> None:
        """Write data to the process

        Args:
            buf: The data to write to the process
        """

        self.in_queue.put(buf)

    def read(self) -> None:
        buf = bytearray()
        while True:
            try:
                buf += self.out_queue.get(block=False)
            except queue.Empty:
                break

        if self.t_out.is_alive() and not buf:
            return None
        return bytes(buf)

    def start(self, *_) -> None:
        """Start the process and the input/output threads"""

        self.p = subprocess.Popen(
            self.cmd,
            cwd=self.cwd,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        self.base.logger.info(f"PID: {self.p.pid} CMD: {self.cmd} CWD: {self.cwd}")

        Thread(target=self._process_in, daemon=True).start()
        self.t_out = Thread(target=self._process_out, daemon=True)
        self.t_out.start()

        # Debugging purposes
        self.t_err = Thread(target=self._process_err, daemon=True)
        self.t_err.start()

    def stop(self, *_) -> None:
        """Stop the process"""

        self.alive = False
        self.p.kill()
        self.p.wait()

    def _process_in(self) -> None:
        while self.alive:
            try:
                chunk = self.in_queue.get(timeout=5)
                self.p.stdin.write(chunk)
            except queue.Empty:
                continue
            except OSError:
                continue

            try:
                self.p.stdin.flush()
            except OSError:
                pass

    def _process_out(self) -> None:
        while self.alive:
            data = self.p.stdout.read(1)
            if not data:
                break
            self.out_queue.put(data)

    def _process_err(self) -> None:
        while self.alive:
            data = self.p.stderr.read(1)
            if not data:
                break
            print(
                data.decode(errors="ignore"), end="", flush=True
            )  # Print to the console



import tkinter as tk

from biscuit.common.icons import Icons

from ..sidebar_item import SideBarViewItem
from .item import ChangeItem


class Changes(SideBarViewItem):
    """The Changes view.

    The Changes view allows the user to view the changes made to the active document.
    - Discard changes.
    - Add changes.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        self.__actions__ = (
            (Icons.DISCARD, self.git_discard_all),
            (Icons.ADD, self.git_add_all),
        )
        self.title = "Changes"
        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.views.sidebar.item)
        self.items = {}

    def refresh(self) -> None:
        if not self.items:
            self.itembar.hide_content()
        else:
            self.itembar.show_content()
        # self.update()

    def clear_tree(self, *_) -> None:
        for item, _ in self.items.values():
            try:
                item.pack_forget()
                item.destroy()
            except:
                pass

        self.items.clear()

    def clear(self, otherthan: list = []) -> None:
        if not otherthan:
            return self.clear_tree()

        for path, val in list(self.items.items()):
            if (path, val[1]) not in otherthan:
                self.items[path][0].destroy()
                del self.items[path]

    def add_item(self, path, kind) -> None:
        if path in self.items and self.items[path][1] == kind:
            return

        new_item = ChangeItem(self, path, kind)
        new_item.pack(fill=tk.X, in_=self.content)
        self.items[path] = (new_item, kind)

    def git_add_all(self, *_) -> None:
        if unstaged := [(path, item[1]) for path, item in self.items.items()]:
            self.base.git.repo.stage_files(*unstaged)
            self.master.open_repo()

    def git_discard_all(self, *_) -> None:
        self.base.git.repo.discard_changes(self.path)
        self.master.master.open_repo()



"""Various widgets/functions used across the editor"""

from .bubble import Bubble
from .buttons import *
from .buttonsentry import ButtonsEntry
from .closables import Closable
from .hintedentry import Entry
from .icon import *
from .labels import *
from .linklabel import *
from .native import *
from .scrollableframe import ScrollableFrame
from .scrollbar import Scrollbar
from .shortcut import Shortcut
from .tree import Tree


@staticmethod
def clamp(value, min_val, max_val):
    return min(max(min_val, value), max_val)



import tkinter as tk

from biscuit.common.icons import Icons
from biscuit.common.ui import Entry

from ..sidebar_view import SideBarView
from .results import Results


class Extensions(SideBarView):
    """View that displays the installed extensions.

    The Extensions view displays the installed extensions.
    #TODO search for extensions
    """

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.__icon__ = Icons.EXTENSIONS
        self.name = "Extensions"

        self.searchterm = tk.StringVar(self)

        self.searchbox = Entry(self, hint="Search extensions")
        self.searchbox.pack(fill=tk.X, anchor=tk.N, padx=10, pady=7)

        self.results = Results(self)
        self.add_item(self.results)
        self.add_action(Icons.FILTER, self.results.toggle_installed)
        self.add_action(Icons.REFRESH, self.results.refresh)
        self.add_action(Icons.SEARCH, self.results.search)
        self.searchbox.bind('<Return>', lambda e:self.results.search()) # Bind ENTER KEY PRESS TO Search Function
        # Error faced & Reason for using Lambda:
        # # https://stackoverflow.com/questions/23842770/python-function-takes-1-positional-argument-but-2-were-given-how

        # self.add_action(Icons.CLEAR_ALL, self.results.clear)

    def initialize(self) -> None:
        self.results.refresh()
        self.results.gui_refresh_loop()



import tkinter as tk

from biscuit.common.icons import Icons
from biscuit.common.ui import Entry, Frame, IconLabelButton, ScrollableFrame

from .changes import Changes
from .placeholder import ChangesTreePlaceholder
from .stagedchanges import StagedChanges


class Git(Frame):
    """The Git view.

    The Git view allows the user to manage the source control of the active document.
    - Show changes.
    - Show staged changes.
    - Commit changes.
    - Push changes.
    - Pull changes.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.views.sidebar.item)

        self.commitbox = Frame(self, **self.base.theme.views.sidebar.item)
        self.commit_message = Entry(
            self.commitbox, hint="Message", **self.base.theme.utils.entry
        )
        self.commit_message.pack(fill=tk.X, pady=(0, 5))

        self.commit_button = IconLabelButton(
            self.commitbox,
            text="Commit",
            icon=Icons.CHECK,
            callback=self.commit,
            highlighted=True,
        )
        self.commit_button.pack(fill=tk.BOTH, side=tk.LEFT, expand=True)

        # Commit menu
        # tk.Label(self.commitbox, text="｜", **self.base.theme.utils.colorlabel).pack(side=tk.LEFT, fill=tk.Y)
        # self.more = IconButton(self.commitbox, icon='chevron-down')
        # self.more.config(**self.base.theme.utils.button)
        # self.more.pack(fill=tk.BOTH)

        self.container = ScrollableFrame(self, **self.base.theme.views.sidebar.item)
        self.container.canvas.config(**self.base.theme.views.sidebar.item)

        self.staged_changes_tree = StagedChanges(self, *args, **kwargs)
        self.container.add(self.staged_changes_tree, fill=tk.BOTH, expand=True)
        self.changes_tree = Changes(self, *args, **kwargs)
        self.container.add(self.changes_tree, fill=tk.BOTH, expand=True)

        self.placeholder = ChangesTreePlaceholder(self)
        self.placeholder.pack(fill=tk.BOTH, expand=True)

    def add_staged_changes(self, changed_files=(), kind=0) -> None:
        for file in changed_files:
            self.staged_changes_tree.add_item(file, kind)

        self.staged_changes_tree.refresh()

    def add_changes(self, changed_files=(), kind=0) -> None:
        for file in changed_files:
            self.changes_tree.add_item(file, kind)

        self.changes_tree.refresh()

    def open_repo(self) -> None:
        # self.staged_changes_tree.clear_tree()
        # self.changes_tree.clear_tree()
        # self.set_title(f"{os.path.basename(self.base.active_directory)}({self.base.git.active_branch})")

        if not self.base.git.repo:
            return

        staged = []

        if deleted := self.base.git.repo.get_staged_deleted_files():
            self.add_staged_changes(deleted, 0)
            staged += deleted
        if added := self.base.git.repo.get_staged_added_files():
            self.add_staged_changes(added, 1)
            staged += added
        if modified := self.base.git.repo.get_staged_modified_files():
            self.add_staged_changes(modified, 2)
            staged += modified

        if staged:
            self.staged_changes_tree.clear(otherthan=staged)

        unstaged = []

        if deleted := self.base.git.repo.get_deleted_files():
            self.add_changes(deleted, 0)
            unstaged += [(i, 0) for i in deleted]
        if added := self.base.git.repo.get_added_files():
            self.add_changes(added, 1)
            unstaged += [(i, 1) for i in added]
        if modified := self.base.git.repo.get_modified_files():
            self.add_changes(modified, 2)
            unstaged += [(i, 2) for i in modified]

        # untracked files
        if untracked := self.base.git.repo.get_untracked_files():
            self.add_changes(untracked, 3)
            unstaged += [(i, 3) for i in untracked]

        if unstaged:
            self.changes_tree.clear(otherthan=unstaged)

    def toggle_staged(self, *_) -> None:
        if not self.base.git_found:
            return

        if self.staged_changes_tree.winfo_ismapped():
            self.staged_changes_tree.pack_forget()
        else:
            self.staged_changes_tree.pack(
                fill=tk.BOTH, before=self.changes_tree, in_=self.container
            )

    def toggle_changes(self, *_) -> None:
        if not self.base.git_found:
            return

        if self.changes_tree.winfo_ismapped():
            self.changes_tree.pack_forget()
        else:
            self.changes_tree.pack(
                fill=tk.BOTH, after=self.staged_changes_tree, in_=self.container
            )

    def enable_tree(self) -> None:
        self.placeholder.pack_forget()
        self.commitbox.pack(padx=(15, 10), pady=5, fill=tk.BOTH)
        self.container.pack(fill=tk.BOTH, expand=True)

        self.staged_changes_tree.clear_tree()
        self.changes_tree.clear_tree()
        self.open_repo()

    def disable_tree(self) -> None:
        self.commitbox.pack_forget()
        self.container.pack_forget()
        self.placeholder.pack(fill=tk.BOTH, expand=True)

    def get_commit_message(self) -> str:
        return self.commit_message.get()

    def commit(self, *_) -> None:
        self.base.git.repo.commit_files(self.get_commit_message())

    def push(self, *_) -> None:
        try:
            self.base.git.repo.push_files()
        except AttributeError:
            self.base.logger.error("No git repository found.")

    def pull(self, *_) -> None:
        try:
            self.base.git.repo.pull_files()
        except AttributeError:
            self.base.logger.error("No git repository found.")



import os
import tkinter as tk

from .breadcrumbs.breadcrumbs import BreadCrumbs
from .editorbase import BaseEditor


class Editor(BaseEditor):
    """Editor widget

    This is the main editor widget that is used to open and edit any type of file.
    Such as text, markdown, html, image, git diff etc. It provides breadcrumbs for opened file.

    This widget acts as a container for the actual editor content.
    `Editor.content` is the actual underlying BaseEditor instance that is used to open the file.
    """

    def __init__(
        self,
        master,
        path: str = None,
        exists: bool = False,
        path2: str = None,
        diff: bool = False,
        language: str = None,
        load_file: bool = True,
        standalone: bool = False,
        config_file: str = None,
        showpath: bool = True,
        preview_file_callback=None,
        open_file_callback=None,
        *args,
        **kwargs,
    ) -> None:
        """Editor widget

        Args:
            master: The parent widget
            path (str): The path of the file to be opened
            exists (bool): Whether the file exists
            path2 (str): The path of the file to be opened in diff, required if diff=True is passed
            diff (bool): Whether the file is to be opened in diff editor
            language (str): Use the `Languages` enum provided (eg. Languages.PYTHON, Languages.TYPESCRIPT)
                This is given priority while picking suitable highlighter. If not passed, guesses from file extension.
            config_file (str): path to the custom config (TOML) file, uses theme defaults if not passed
            showpath (bool): Whether to show the breadcrumbs for editor or not
            preview_file_callback (function): called when files in breadcrumbs-pathview are single clicked. MUST take an argument (path)
            open_file_callback (function): called when files in breadcrumbs-pathview are double clicked. MUST take an argument (path)
        """

        super().__init__(master, *args, **kwargs)

        self.path = path
        self.exists = exists
        self.path2 = path2
        self.diff = diff
        self.language = language
        self.standalone = standalone
        self.showpath = showpath
        self.config_file = config_file
        self.preview_file_callback = preview_file_callback
        self.open_file_callback = open_file_callback

        self.config(bg=self.base.theme.border)
        self.grid_columnconfigure(0, weight=1)

        self.content = self.base.editortypes.get_editor(
            self,
            path,
            exists,
            path2,
            diff,
            language,
            load_file=load_file,
            standalone=standalone,
        )
        self.filename = os.path.basename(self.path) if path else None

        self.grid_rowconfigure(0, weight=1)
        self.content.grid(row=0, column=0, sticky=tk.NSEW)

    def save(self, path: str = None) -> None:
        """Save the content to the file

        Args:
            path (str, Optional): The path to save the content to. If not passed, uses the current path
        """

        self.content.save(path)

    def focus(self) -> None:
        """Focus the editor content"""

        self.content.focus()

    def __str__(self) -> str:
        return f"{self.content.__class__.__name__}({self.path})"



import tkinter as tk

from hintedtext import HintedEntry

from biscuit.common.icons import Icons
from biscuit.common.ui import Entry, Frame, Scrollbar

from ..panelview import PanelView

HELP = """
Welcome to Biscuit Inspect Console
----------------------------------

This is a Python console that allows you to interact with the Biscuit at runtime.
You can use this console to run commands, check the status of the application, and more.
`self` refers to the App instance, and you can access all the attributes and methods of the App class.

example usage:
- self.open_editor("path/to/file", False).content.text.write("Hello, World!")

Custom commands:
- help: Show this help message
- clear: Clear the console
"""


class Inspect(PanelView):
    """The Inspect view."""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.__actions__ = ((Icons.CLEAR_ALL, self.clear),)

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        container = Frame(self)
        container.grid(row=0, column=0, sticky=tk.NSEW)

        container.rowconfigure(0, weight=1)
        container.columnconfigure(0, weight=1)

        self.text = tk.Text(
            container,
            relief=tk.FLAT,
            padx=10,
            pady=10,
            font=("Consolas", 11),
            **self.base.theme.editors.text,
        )
        self.text.grid(row=0, column=0, sticky=tk.NSEW)

        self.scrollbar = Scrollbar(container, style="EditorScrollbar")
        self.scrollbar.grid(sticky=tk.NSEW, row=0, column=1)

        frame = Frame(
            self,
            padx=5,
            pady=5,
            bg=self.base.theme.layout.background,
        )
        frame.grid(row=1, column=0, sticky=tk.NSEW, pady=(1, 0))

        self.entry = HintedEntry(
            frame,
            relief=tk.FLAT,
            font=("Consolas", 11),
            hint="Use /help for help",
            bg=self.base.theme.layout.background,
            fg=self.base.theme.layout.foreground,
        )
        self.entry.pack(side=tk.LEFT, fill=tk.X, expand=True)
        self.entry.bind("<Return>", self.enter)
        self.entry.bind("<Up>", self.previous_command)
        self.entry.bind("<Down>", self.next_command)

        self.text.tag_config(
            "hline",
            background=self.base.theme.border,
            bgstipple=f"@{self.base.resources.line}",
        )

        self.text.config(yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.text.yview)

        self.history = []
        self.history_index = -1

        self.show_help()
        self.text.config(state=tk.DISABLED)

    def enter(self, *_) -> None:
        text = self.entry.get()
        self.write()
        self.entry.delete(0, tk.END)

        if text.strip():
            self.history.append(text)
            self.history_index = len(self.history)

        match text:
            case "/help":
                self.show_help()
            case "/clear":
                self.clear()

        if result := self.base.control_execute(text):
            self.write(str(result))

        self.show_separator()

    def previous_command(self, _):
        if self.history:
            self.history_index = max(0, self.history_index - 1)
            self.entry.delete(0, tk.END)
            self.entry.insert(0, self.history[self.history_index])

    def next_command(self, _):
        if self.history:
            self.history_index = min(len(self.history), self.history_index + 1)
            self.entry.delete(0, tk.END)
            if self.history_index < len(self.history):
                self.entry.insert(0, self.history[self.history_index])

    def write(self, text="") -> None:
        self.text.config(state=tk.NORMAL)
        self.text.insert(tk.END, f"{text}\n")
        self.text.see(tk.END)
        self.text.config(state=tk.DISABLED)

    def show_help(self) -> None:
        self.write(HELP)

    def show_separator(self) -> None:
        self.text.config(state=tk.NORMAL)
        self.text.insert(tk.END, "\n", "hline")
        self.text.config(state=tk.DISABLED)

    def clear(self, *_) -> None:
        self.text.config(state=tk.NORMAL)
        self.text.delete(1.0, tk.END)
        self.text.config(state=tk.DISABLED)



from __future__ import annotations

import os
import typing

import tarts as lsp

from .client import LangServerClient
from .languages import Languages
from .utils import decode_position

if typing.TYPE_CHECKING:
    from biscuit import App
    from biscuit.editor.text import Text


class LanguageServerManager:
    """Language Server Manager

    This class is used to manage the language server clients. It is responsible for creating, updating, and deleting
    the language server clients. It also manages the requests made by the user to the language server clients.
    """

    def __init__(self, base: App):
        self.base = base

        self.langservers: dict[str, str] = {}

        # built-in support for python-lsp-server
        self.langservers[Languages.PYTHON] = "pylsp"

        self.existing: dict[str, LangServerClient] = {}
        self.latest: LangServerClient = None

        self.kill_thread = None

    def _update_symbols(
        self, tab: Text, resp: typing.List[lsp.SymbolInformation]
    ) -> None:
        try:
            self.base.settings.symbols_actionset.update(
                [
                    (
                        i.name,
                        lambda _, i=i: self.base.goto_location_in_active_editor(
                            decode_position(i.location.range.start)
                        ),
                    )
                    for i in resp
                ]
            )
            self.base.palette.update()
        except:
            # TODO: currently only supports lsp.SymbolInformation; implement this for lsp.DocumentSymbol
            pass

    def register_langserver(self, language: str, command: str) -> None:
        """Register a language server for a specific language

        Args:
            language (str): The language for which the language server is being registered
            command (str): The command to start the language server
        """

        self.langservers[language] = command

    def tab_opened(self, tab: Text) -> None:
        """Let the language server know that a tab has been opened"""

        self.latest = self.request_client_instance(tab)
        if self.latest:
            self.latest.open_tab(tab)
            self.latest.request_outline(tab)
        return self.latest

    def request_removal(self, tab: Text) -> None:
        """Request the language server to remove a tab"""

        for instance in list(self.existing.values()):
            instance.close_tab(tab)

    def tab_closed(self, tab: Text) -> None:
        """Let the language server know that a tab has been closed"""

        if inst := self.request_client_instance(tab):
            inst.close_tab(tab)

    def request_completions(self, tab: Text) -> None:
        """Request completions from the language server for a tab"""

        for instance in list(self.existing.values()):
            if tab in instance.tabs_opened:
                instance.request_completions(tab)

    def request_goto_definition(self, tab: Text) -> None:
        """Request the language server to go to the definition of a symbol in a tab"""

        for instance in list(self.existing.values()):
            if tab in instance.tabs_opened:
                instance.request_go_to_definition(tab)

    def request_references(self, tab: Text) -> None:
        """Request references to a symbol in a tab from the language server"""

        for instance in list(self.existing.values()):
            if tab in instance.tabs_opened:
                instance.request_references(tab)

    def request_rename(self, tab: Text, new_name: str) -> None:
        """Request the language server to rename a symbol in a tab"""

        for instance in list(self.existing.values()):
            if tab in instance.tabs_opened:
                instance.request_rename(tab, new_name)

    def request_hover(self, tab: Text) -> None:
        """Request the language server to provide a hover for a symbol in a tab"""

        for instance in list(self.existing.values()):
            if tab in instance.tabs_opened:
                instance.request_hover(tab)

    def request_outline(self, tab: Text) -> None:
        """Request the language server to provide an outline for a tab"""

        for instance in list(self.existing.values()):
            if tab in instance.tabs_opened:
                instance.request_outline(tab)

    def content_changed(self, tab: Text) -> None:
        """Content of a tab has changed, notify the language server about it"""

        for instance in list(self.existing.values()):
            if tab in instance.tabs_opened:
                instance.send_change_events(tab)

    def request_client_instance(self, tab: Text) -> LangServerClient | None:
        """Request a language server client instance for a specific language and workspace root directory.

        If a client instance already exists for the language and the root directory of the tab, it is returned.
        Otherwise, a new client instance is created and returned.

        Args:
            tab (Text): The tab for which the language server client instance is being requested
        """

        if tab.path is None or (
            tab.language not in self.langservers.keys()
            and tab.language.lower() not in self.langservers.keys()
            and tab.language_alias not in self.langservers.keys()
        ):
            return

        root_dir = self.base.active_directory or os.path.dirname(tab.path)

        try:
            return self.existing[(root_dir, tab.language_alias)]
        except KeyError:
            pass

        self.base.statusbar.process_indicator.show()
        self.base.logger.trace(
            f"<<-- Requesting <LSPC>({tab.language_alias}) instance for --[{root_dir}] -->>"
        )

        # TODO multithread this process
        langserver = LangServerClient(self, tab, root_dir)
        langserver.run_loop()
        self.existing[(root_dir, tab.language_alias)] = langserver

        return langserver

    def kill(self, instance: LangServerClient) -> None:
        """Kill a language server client instance

        Args:
            instance (LangServerClient): The language server client instance to be killed
        """

        if not self.existing.get((instance.root_dir, instance.language), None):
            return

        self.base.logger.trace(
            f"-- Killing LSPC({instance.language}) PID: {instance.io.p.pid} --"
        )

        self.existing.pop((instance.root_dir, instance.language))
        if instance.client.state == lsp.ClientState.NORMAL:
            instance.client.shutdown()
        else:
            instance.io.p.kill()

        del instance



from typing import Callable, List, Tuple


class ActionSet(list):
    """Action Set
    Actions are represented by tuples of the form (command: str, callback: Callable).
    Action sets are used to group actions together and display them in the palette.

    The pinned actions are displayed at the top of the palette, their command can be a format string,
    the palette will format the command with the search term.
    eg: pinned=[["Search on Google: {}", foo_method],]
    """

    def __init__(
        self,
        description: str,
        prefix: str,
        items: List[List[str | Callable]] = [],
        pinned: List[List[str | Callable[[str], None]]] = [],
        *args,
        **kwargs
    ) -> None:
        """Initializes the actionset.

        Args:
            description (str): The description of the actionset.
            prefix (str): The prefix of the actionset.
            items (List[Tuple[str | Callable]], optional): The list of actions. Defaults to [].
            pinned (List[List[str | Callable[[str], None]]], optional): The list of pinned actions. Defaults to [].
        """

        super().__init__(items, *args, **kwargs)
        self.description: str = description
        self.prefix: str = prefix

        self.pinned: List[Tuple[str, Callable[[str], None]]] = (
            pinned  # [[command, callback], ...]
        )

    def __repr__(self) -> str:
        return self.description

    def update(self, items: Tuple[str, Callable]) -> None:
        """Update the actionset with new items.

        Args:
            items (Tuple[str, Callable]): The new items to be added."""

        self.clear()
        self += items

    def add_action(self, command: str, callback: Callable) -> None:
        """Add an item to the actionset.

        Args:
            command (str): The command to be displayed.
            callback (Callable): The callback function to be called."""

        self.append((command, callback))

    def add_pinned_actions(self, command: str, callback: Callable) -> None:
        """Add an item to the pinned actions.
        The command can be a format string, the palette will format the command with the search term.

        Args:
            command (str): The command to be displayed.
            callback (Callable): The callback function to be called."""

        self.pinned.append((command, callback))

    def get_pinned(self, term) -> list:
        """Not to be called directly. Returns the pinned actions with the search term formatted."""

        return [[item[0].format(term or "...")] + item[1:] for item in self.pinned]



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common.icons import Icons
from biscuit.common.ui import Entry, Frame, IconLabelButton, WebLinkLabel, WrappingLabel

if typing.TYPE_CHECKING:
    ...


class AIPlaceholder(Frame):
    """Home page for the AI assistant view.

    The AIPlaceholder is displayed when the user has not configured the AI assistant.
    - User can enter their API key to start using the AI assistant.
    - API key can be acquired from the Google AI Studio website.
    - The key is stored in the secrets database."""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(padx=15, pady=10, **self.base.theme.views.sidebar.item)

        self.label = WrappingLabel(
            self,
            font=self.base.settings.uifont,
            justify=tk.LEFT,
            anchor=tk.W,
            **self.base.theme.views.sidebar.item.content,
            text="AI chat to find answer to your queries, configure Google AI API key to start using",
        )
        self.label.pack(fill=tk.X)

        self.api_key = tk.StringVar()
        self.api_entry = Entry(
            self, hint="Enter API Key here...", textvariable=self.api_key
        )
        self.api_entry.pack(fill=tk.X, pady=5)

        confirm_btn = IconLabelButton(
            self,
            text="Start Chat",
            icon=Icons.SPARKLE_FILLED,
            callback=self.start_chat,
            pady=2,
            highlighted=True,
        )
        confirm_btn.pack(fill=tk.X, pady=5)

        self.link = WebLinkLabel(
            self,
            text="Don't have an API key? Get it from here.",
            link="https://aistudio.google.com/app/apikey",
        )
        self.link.config(bg=self.base.theme.views.sidebar.background)
        self.link.pack(fill=tk.X, pady=5)

    def start_chat(self) -> None:
        self.master.add_chat(self.api_key.get())



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.editor.editorbase import BaseEditor

from .shortcuts import Shortcuts

if typing.TYPE_CHECKING:
    from . import EditorsManager


class Placeholder(BaseEditor):
    """Placeholder for the Editors Manager

    - Placeholder for the Editors
    - Shows the shortcuts
    """

    def __init__(self, master: EditorsManager, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(bd=0, relief=tk.FLAT, **self.base.theme.editors)

        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)

        # try:
        #     self.logo = tk.Label(
        #         self,
        #         image=self.base.settings.resources.logo,
        #         **self.base.theme.editors.labels,
        #     )
        #     self.logo.grid(row=0, column=0)
        # except tk.TclError:
        #     pass

        self.shortcuts = Shortcuts(self, **self.base.theme.editors)
        self.shortcuts.grid(row=0, column=0)

        self.shortcuts.add_shortcut("Show all commands", ["Ctrl", "Shift", "P"])
        self.shortcuts.add_shortcut("Toggle terminal", ["Ctrl", "`"])
        self.shortcuts.add_shortcut("Open Folder", ["Ctrl", "Shift", "O"])
        self.shortcuts.add_shortcut("View Extensions", ["Ctrl", "Shift", "X"])
        self.shortcuts.add_shortcut("Open Settings", ["Ctrl", ","])

        self.bind("<Double-Button-1>", self.base.commands.new_file)



import tkinter as tk
import typing

if typing.TYPE_CHECKING:
    from biscuit import App


class Text(tk.Text):
    """Text widget with reference to base"""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.base: App = master.base


class Toplevel(tk.Toplevel):
    """Custom Toplevel widget for the app."""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.base: App = master.base


class Canvas(tk.Canvas):
    """Canvas with reference to base"""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.base: App = master.base


class Menubutton(tk.Menubutton):
    """Menubutton with reference to base"""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.base: App = master.base


class Label(tk.Label):
    """Label with reference to base"""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.base: App = master.base

    def set_text(self, text: str) -> None:
        self.config(text=text)


class Frame(tk.Frame):
    """Frame with reference to base"""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.base: App = master.base



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common.ui import Frame
from biscuit.views import *

if typing.TYPE_CHECKING:
    from biscuit.layout.statusbar.activitybar import ActivityBar


class SecondarySideBar(Frame):
    """Secondary side bar of the application

    - Contains the SidebarViews
    - Manages the SidebarViews
    """

    def __init__(
        self, master: Frame, activitybar: ActivityBar, *args, **kwargs
    ) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.border)

        self.rowconfigure(1, weight=1)
        self.columnconfigure(1, weight=1)

        self.views = []
        self.active_view = None
        self.visible = False

        self.secondary_activitybar = activitybar
        self.secondary_activitybar.attach_sidebar(self)

        self.default_views = {
            "Outline": Outline(self),
            "GitHub": GitHub(self),
            "Extensions": Extensions(self),
            "AI": AI(self),
            "Source Control": SourceControl(self),
        }
        self.add_views(self.default_views.values())

    def toggle(self) -> None:
        """Toggle the sidebar."""

        if self.visible:
            self.hide()
        else:
            if not self.active_view:
                self.show_source_control()
            else:
                self.pack()

    def add_views(self, views: list[SideBarView]) -> None:
        """Adds multiple views to the sidebar at once."""

        for view in views:
            self.add_view(view)

    def add_view(self, view: SideBarView) -> None:
        """Adds a view to the sidebar.

        Args:
            view (SidebarView): The view to add."""

        self.views.append(view)
        self.secondary_activitybar.add_view(view)

    def create_view(self, name: str, icon: str = "browser") -> SideBarView:
        """Create a new view.

        Args:
            name (str): The name of the view.
            icon (str, optional): The icon of the view. Defaults to "browser"
        """

        view = SideBarView(self, name, icon)
        self.add_view(view)
        return view

    def delete_all_views(self) -> None:
        """Permanently delete all views."""

        for view in self.views:
            view.destroy()

        self.views.clear()

    def delete_view(self, view: SideBarView) -> None:
        """Permanently delete a view.

        Args:
            view (SidebarView): The view to delete."""

        view.destroy()
        self.views.remove(view)

    @property
    def outline(self) -> Outline:
        return self.default_views["Outline"]

    @property
    def source_control(self) -> SourceControl:
        return self.default_views["Source Control"]

    @property
    def ai(self) -> AI:
        return self.default_views["AI"]

    @property
    def github(self) -> GitHub:
        return self.default_views["GitHub"]

    @property
    def extensions(self) -> Extensions:
        return self.default_views["Extensions"]

    def show_view(self, view: SideBarView) -> SideBarView:
        """Show a view in the sidebar.

        Args:
            view (SidebarView): The view to show."""

        for i in self.secondary_activitybar.buttons:
            if i.view == view:
                self.secondary_activitybar.set_active_slot(i)
                i.enable()
                self.active_view = view
                return view

    def show_explorer(self, *_) -> Explorer:
        return self.show_view(self.explorer)

    def show_outline(self, *_) -> Outline:
        return self.show_view(self.outline)

    def show_search(self, *_) -> Search:
        return self.show_view(self.search)

    def show_source_control(self, *_) -> SourceControl:
        return self.show_view(self.source_control)

    def show_debug(self, *_) -> Debug:
        return self.show_view(self.debug)

    def show_ai(self, *_) -> AI:
        return self.show_view(self.ai)

    def show_github(self, *_) -> GitHub:
        return self.show_view(self.github)

    def show_extensions(self, *_) -> Extensions:
        return self.show_view(self.extensions)

    def pack(self):
        super().pack(side=tk.LEFT, fill=tk.Y, after=self.base.contentpane, padx=(1, 0))
        self.visible = True

    def hide(self):
        super().pack_forget()
        self.visible = False



import json

import requests


class Releases:
    """Releases API

    This class is used to interact with the GitHub Releases API.
    Provides methods to get the latest release, the latest release for a specific OS
    or a release by tag."""

    def __init__(self, base=None):
        self.base = base

    def get(self, url: str):
        resp = requests.get(url)
        if resp.status_code == 200:
            return json.loads(resp.text)

    def get_latest_release(self, owner: str, repo: str):
        return self.get(f"https://api.github.com/repos/{owner}/{repo}/releases/latest")

    def get_release(self, owner: str, repo: str, tag: str):
        return self.get(
            f"https://api.github.com/repos/{owner}/{repo}/releases/tags/{tag}"
        )

    def get_latest_windows_release(self, owner: str, repo: str):
        resp = self.get_latest_release(owner, repo)
        for asset in resp["assets"]:
            if "-windows-" in asset["name"].lower():
                return asset["browser_download_url"]

    def get_latest_macos_release(self, owner: str, repo: str):
        resp = self.get_latest_release(owner, repo)
        for asset in resp["assets"]:
            if "-mac-" in asset["name"].lower():
                return asset["browser_download_url"]

    def get_latest_linux_release(self, owner: str, repo: str):
        resp = self.get_latest_release(owner, repo)
        for asset in resp["assets"]:
            if "-linux-" in asset["name"].lower():
                return asset["browser_download_url"]

    def get_windows_release(self, owner: str, repo: str, tag: str):
        resp = self.get_release(owner, repo, tag)
        for asset in resp["assets"]:
            if "-windows-" in asset["name"].lower():
                return asset["browser_download_url"]

    def get_macos_release(self, owner: str, repo: str, tag: str):
        resp = self.get_release(owner, repo, tag)
        for asset in resp["assets"]:
            if "-mac-" in asset["name"].lower():
                return asset["browser_download_url"]

    def get_linux_release(self, owner: str, repo: str, tag: str):
        resp = self.get_release(owner, repo, tag)
        for asset in resp["assets"]:
            if "-linux-" in asset["name"].lower():
                return asset["browser_download_url"]



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from biscuit.api import ExtensionsAPI


class Extension:
    """Biscuit Extension base class.

    This class can be optionally inherited to create an extension."""

    def __init__(self, api: ExtensionsAPI) -> None:
        self.api = api

    def install(self) -> None:
        """Install the extension."""
        ...

    def uninstall(self) -> None:
        """Uninstall the extension."""
        ...


def setup(api: ExtensionsAPI) -> Extension:
    """Just for type hinting."""
    ...



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, IconButton

if typing.TYPE_CHECKING:
    from .debug import Debug


class DebuggerActions(Frame):
    def __init__(self, debug: Debug):
        super().__init__(debug)
        self.debug = debug
        self.base = debug.base
        self.manager = debug.base.debugger_manager
        self.config(**self.base.theme.views.sidebar)

        self.pause_btn = self.add_button(
            Icons.DEBUG_PAUSE, self.toggle_pause, "#72bdff", icon2=Icons.DEBUG_CONTINUE
        )

        self.actions = {
            Icons.DEBUG_STEP_OVER: (self.step_over, "#72bdff"),
            Icons.DEBUG_STEP_INTO: (self.step_into, "#72bdff"),
            Icons.DEBUG_STEP_OUT: (self.step_out, "#72bdff"),
            Icons.DEBUG_RESTART: (self.restart, "#87d282"),
            Icons.DEBUG_STOP: (self.stop, "#f6876d"),
        }

        for icon, props in self.actions.items():
            action, color = props
            self.add_button(icon, action, color)

    def add_button(
        self, icon: str, action: typing.Callable, color: str, icon2: str = None
    ):
        button = IconButton(self, icon, action, icon2=icon2)
        button.config(fg=color, activeforeground=color)
        button.pack(side=tk.LEFT)
        return button

    def toggle_pause(self, *_):
        self.base.commands.debugger_toggle_pause()

    def step_over(self, *_):
        self.base.commands.debugger_step_over()

    def step_into(self, *_):
        self.base.commands.debugger_step_into()

    def step_out(self, *_):
        self.base.commands.debugger_step_out()

    def restart(self, *_):
        self.base.commands.debugger_restart()

    def stop(self, *_):
        self.base.commands.debugger_stop()



from __future__ import annotations

import typing

import click

from biscuit.git import URL

if typing.TYPE_CHECKING:
    from biscuit import App


@click.command()
@click.argument(
    "path",
    type=click.Path(
        exists=True,
        dir_okay=True,
        resolve_path=True,
    ),
    required=False,
)
def open(path=None) -> typing.Callable[[App, str], None]:
    """Open a file or folder in the editor

    This command will open a file or folder in the editor.

    Example:
        biscuit open path/to/file

    Args:
        path (str, optional): The path to the file or folder. Defaults to None.
    """

    return lambda app, path=path: app.open(path)


@click.command()
@click.argument(
    "path",
    type=click.Path(
        exists=True,
        dir_okay=False,
        resolve_path=True,
    ),
    required=False,
)
@click.argument("linecol", type=str, required=False)
def goto(path=None, linecol=None) -> typing.Callable[[App, str], None]:
    """Open a file and go to a specific location

    This command will open a file and go to a specific location.

    Example:
        biscuit goto path/to/file line \n
        biscuit goto path/to/file line:column

    Args:
        path (str, optional): The path to the file. Defaults to None.
        linecol (str, optional): The line and column to go to. Defaults to None.
    """

    if not path:
        path = click.prompt("path/to/file", type=str)
    if not linecol:
        linecol = click.prompt("line:column", type=str)
    if not linecol:
        linecol = "1:1"

    # TODO: make the column optional

    return lambda app, path=path, linecol=linecol: app.goto_location(
        path, linecol.replace(":", ".")
    )


def register(cli: click.Group):
    cli.add_command(goto)
    cli.add_command(open)



"""Helper functions for the core components."""

import inspect
import os
import subprocess as sp
import webbrowser
from tkinter import messagebox

import filetype


def show_python_not_installed_message():
    messagebox.showerror(
        "Python not found",
        "Python is not installed on your system/not in PATH.\n\nBiscuit require Python to be installed. Please install Python and try again.",
    )
    webbrowser.open("https://www.python.org/downloads/")
    exit(1)


def check_python_installation():
    """Check if python is installed and install python language server."""

    try:
        if os.name == "nt":
            sp.check_call(["python", "--version"])
            sp.Popen(["pip", "install", "python-lsp-server"])
        else:
            sp.check_call(["python3", "--version"])
            sp.Popen(["python3", "-m", "pip", "install", "python-lsp-server"])

    except sp.CalledProcessError:
        show_python_not_installed_message()


def get_file_type(file_path):
    """Get the type of the file.

    Args:
        file_path (str): path to the file
    """

    return filetype.guess(file_path)


def is_image(file_path):
    """Check if the file is an image.

    Args:
        file_path (str): path to the file
    """

    return filetype.is_image(file_path)


def search_google(query: str) -> None:
    webbrowser.open(f"https://www.google.com/search?q={query}")


def caller_name(skip=2):
    """
    Get the name of a caller in the format module.class.method

    `skip` specifies how many levels of stack to skip while getting caller
    name. skip=1 means "who calls me", skip=2 "who calls my caller" etc.

    An empty string is returned if skipped levels exceed stack height
    """

    stack = inspect.stack()
    start = 0 + skip
    if len(stack) < start + 1:
        return ""
    parentframe = stack[start][0]

    name = []
    module = inspect.getmodule(parentframe)
    # modname can be None when frame is executed directly in console
    if module:
        name.append(module.__name__)
    # detect classname
    if "self" in parentframe.f_locals:
        name.append(parentframe.f_locals["self"].__class__.__name__)

    del parentframe, stack

    return ".".join(name)


def caller_class_name(skip=2):
    """
    Get the name of the class of the caller.

    `skip` specifies how many levels of stack to skip while getting the caller's class.
    skip=1 means "who calls me", skip=2 "who calls my caller" etc.

    An empty string is returned if skipped levels exceed the stack height.
    """
    stack = inspect.stack()
    start = 0 + skip
    if len(stack) < start + 1:
        return ""

    parentframe = stack[start][0]
    class_name = None

    # detect classname
    if "self" in parentframe.f_locals:
        class_name = parentframe.f_locals["self"].__class__.__name__

    del parentframe, stack

    return class_name



import tkinter as tk

from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, Icon, IconButton, Label


class Tab(Frame):
    """Tab for terminal panel."""

    def __init__(self, master, terminal, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.terminal = terminal
        self.selected = False

        self.hbg, self.hfg, self.bg, self.fg = (
            self.base.theme.layout.content.editors.bar.tab.values()
        )
        self.config(bg=self.bg)

        self.icon = Icon(
            self,
            icon=terminal.icon or Icons.TERMINAL,
            width=3,
            **self.base.theme.layout.content.editors.bar.tab.icon
        )
        self.icon.pack(side=tk.LEFT)

        self.name = Label(
            self,
            text=terminal.name or terminal.__class__.__name__,
            padx=5,
            font=self.base.settings.uifont,
            anchor=tk.W,
            bg=self.bg,
            fg=self.fg,
        )
        self.name.pack(side=tk.LEFT, expand=True, fill=tk.X)

        self.closebtn = IconButton(
            self,
            Icons.CLOSE,
            event=self.close,
            **self.base.theme.layout.content.editors.bar.tab.icon
        )
        self.closebtn.pack(padx=(0, 5))

        self.bind("<Button-1>", self.select)
        self.icon.bind("<Button-1>", self.select)
        self.name.bind("<Button-1>", self.select)

        self.bind("<Enter>", self.on_hover)
        self.bind("<Leave>", self.off_hover)

    def close(self, *_) -> None:
        self.master.close_tab(self)

    def on_hover(self, *_) -> None:
        if not self.selected:
            self.icon.config(bg=self.hbg)
            self.name.config(bg=self.hbg)
            self.config(bg=self.hbg)
            self.closebtn.config(bg=self.hbg)

            self.hovered = True

    def off_hover(self, *_) -> None:
        if not self.selected:
            self.icon.config(bg=self.bg)
            self.name.config(bg=self.bg)
            self.config(bg=self.bg)
            self.closebtn.config(bg=self.bg)
            self.hovered = False

    def deselect(self, *_) -> None:
        if self.selected:
            self.terminal.grid_remove()

            self.icon.config(bg=self.bg, fg=self.fg)
            self.name.config(bg=self.bg, fg=self.fg)
            self.config(bg=self.bg)
            self.closebtn.config(bg=self.bg, activeforeground=self.fg)
            self.selected = False

    def select(self, *_) -> None:
        if not self.selected:
            self.master.set_active_tab(self)
            self.terminal.grid(column=0, row=0, sticky=tk.NSEW)

            self.icon.config(bg=self.hbg, fg=self.hfg)
            self.name.config(bg=self.hbg, fg=self.hfg)
            self.config(bg=self.hbg)
            self.closebtn.config(bg=self.hbg, activeforeground=self.hfg)
            self.selected = True



import os
import tkinter as tk

from biscuit.common.icons import Icons
from biscuit.common.ui import Bubble, Frame, IconButton, Label, Menubutton

KINDS = [
    ("D", "Deleted", "red"),
    ("A", "Added", "green"),
    ("M", "Modified", "orange"),
    ("U", "Untracked", "green"),
]


class ChangeItem(Frame):
    """
    Changes tree item.
    Kinds:
        0 - deleted
        1 - added
        2 - modified
        3 - untracked
    """

    def __init__(self, master, path, kind, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.views.sidebar.item)
        self.path = path
        self.kind = kind

        name_label = Menubutton(
            self,
            text=os.path.basename(path),
            anchor=tk.W,
            font=self.base.settings.uifont,
            padx=10,
            pady=2,
            **self.base.theme.views.sidebar.item.button,
        )
        name_label.pack(fill=tk.BOTH, expand=True, side=tk.LEFT)
        name_label.bind("<Double-Button-1>", self.open_diff)

        # path_label = TruncatedLabel(self, text=path, anchor=tk.W, font=("Segoe UI", 9),
        #     padx=10, pady=2, fg="grey", **self.base.theme.views.sidebar.item
        # )
        # path_label.pack(fill=tk.BOTH, side=tk.LEFT)
        # path_label.bind("<Double-Button-1>", self.open_diff)

        IconButton(
            self,
            Icons.DISCARD,
            self.git_discard,
            **self.base.theme.views.sidebar.item.button,
        ).pack(fill=tk.BOTH, side=tk.LEFT)
        IconButton(
            self, Icons.ADD, self.git_add, **self.base.theme.views.sidebar.item.button
        ).pack(fill=tk.BOTH, side=tk.LEFT)
        Label(
            self,
            text=KINDS[self.kind][0],
            fg=KINDS[self.kind][2],
            font=self.base.settings.uifont_bold,
            width=3,
            pady=2,
            **self.base.theme.views.sidebar.item,
        ).pack(fill=tk.BOTH, expand=True)

        self.bubble = Bubble(self, text=f"{path} • {KINDS[self.kind][1]}")
        self.bind("<Enter>", self.bubble.show)
        self.bind("<Leave>", self.bubble.hide)

    def open_diff(self, _) -> None:
        self.base.open_diff(self.path, self.kind)

    def git_add(self, *_) -> None:
        self.base.git.repo.stage_files((self.path, self.kind))
        self.master.master.open_repo()

    def git_discard(self, *_) -> None:
        self.base.git.repo.discard_changes(self.path)
        self.master.master.open_repo()



import tkinter as tk

from hintedtext import HintedEntry

from .native import Frame


class Entry(Frame):
    def __init__(self, master, hint="", *args, **kwargs) -> None:
        super().__init__(master)
        self.config(padx=1, pady=1, bg=self.base.theme.border)
        self.grid_columnconfigure(0, weight=1)

        self.entry = HintedEntry(
            self,
            relief=tk.FLAT,
            font=self.base.settings.uifont,
            **self.base.theme.utils.entry,
            bd=5,
            hint=hint
        )
        self.entry.config(*args, **kwargs)
        self.entry.grid(row=0, column=0, sticky=tk.NSEW)
        self.entry.add_hint(hint)

    def insert(self, *args) -> None:
        self.entry.insert(*args)

    def get(self, *args) -> str:
        return self.entry.get(*args)

    def delete(self, *args) -> None:
        self.entry.delete(*args)

    def bind(self, *args, **kwargs) -> None:
        self.entry.bind(*args, **kwargs)



from __future__ import annotations

import os
import tkinter as tk
import typing
from tkinter.messagebox import askyesno
from typing import Dict, List, Union

from biscuit.common import ActionSet, Game
from biscuit.common.ui import Frame
from biscuit.editor import Editor, Welcome

from .editorsbar import EditorsBar
from .placeholder import Placeholder

if typing.TYPE_CHECKING:
    from biscuit.editor import BaseEditor

    from ..content import Content


class EditorsManager(Frame):
    """Editors Pane

    - Contains the Editorsbar
    - Manages the Editorsbar
    - Manages the Editors
    """

    def __init__(self, master: Content, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.border)

        self.grid_propagate(False)
        self.grid_rowconfigure(1, weight=1)
        self.grid_columnconfigure(0, weight=1)

        self.editorsbar = EditorsBar(self)
        self.editorsbar.grid(row=0, column=0, sticky=tk.EW, pady=(0, 1))

        self.active_editors: List[Editor] = []
        self.closed_editors: Dict[Editor] = {}

        self.closed_editors: List[Editor] = []
        self.max_closed_editors = 10

        self.emptytab = Placeholder(self)
        self.emptytab.grid(column=0, row=1, sticky=tk.NSEW)

        self.default_editors: List[Editor] = [Welcome(self)]

    def is_empty(self) -> bool:
        return not self.active_editors

    def is_open(self, path: str) -> bool:
        return any(
            editor.path
            and path
            and (os.path.abspath(editor.path) == os.path.abspath(path))
            for editor in self.active_editors
        )

    def get_active_actionset(self) -> ActionSet:
        self.actionset.update(
            [(editor.filename, editor) for editor in self.active_editors]
        )
        return self.actionset

    def generate_actionsets(self) -> None:
        self.actionset = ActionSet("Switch to active editors", "active:", [])
        self.base.palette.register_actionset(self.get_active_actionset)

        self.base.palette.register_actionset(
            lambda: ActionSet(
                "Configure Run Command",
                "runconf:",
                pinned=[
                    [
                        "Run: {}",
                        lambda command=None, e=self.base.editorsmanager.active_editor: (
                            e.content.set_run_command(command)
                            if command
                            else print("Command can't be empty!")
                        ),
                    ]
                ],
            )
        )

    def add_default_editors(self) -> None:
        self.add_editors(self.default_editors)

    def add_welcome(self) -> None:
        self.add_editor(Welcome(self))

    def add_editors(self, editors: list[Editor]) -> None:
        for editor in editors:
            self.add_editor(editor)

    def add_editor(self, editor: Union[Editor, BaseEditor]) -> Editor | BaseEditor:
        """Add a new editor to the editor pane.

        Args:
            editor (Union[Editor, BaseEditor]): The editor to add."""

        if editor in self.active_editors:
            return self.set_active_editor(editor)

        self.active_editors.append(editor)
        if editor.content:
            editor.content.create_buttons(self.editorsbar.action_container)
        self.editorsbar.add_tab(editor)
        self.base.open_editors.add_item(editor)
        self.refresh()
        return editor

    def delete_all_editors(self) -> None:
        for tab in self.editorsbar.active_tabs:
            if e := tab.editor:
                self.editorsbar.save_unsaved_changes(e)
                e.destroy()

        self.editorsbar.clear_all_tabs()
        self.active_editors.clear()
        self.base.open_editors.clear()
        self.refresh()

    def close_all_editors(self) -> None:
        for tab in self.editorsbar.active_tabs:
            if e := tab.editor:
                self.editorsbar.save_unsaved_changes(e)
                self.close_editor(e)
        self.refresh()

    def reopen_active_editor(self, *_) -> None:
        if self.active_editor and self.active_editor.exists:
            self.delete_editor(self.active_editor)
            self.update()
            self.open_editor(self.active_editor.path)

    def reopen_editor(self, path: str):
        if not askyesno(
            "Reopen Editor",
            f"You will lose any unsaved changes to ({path}). Are you sure?",
        ):
            return

        try:
            self.delete_editor(self.get_editor(path))
            self.update()
            self.open_editor(path)
        except Exception as e:
            self.base.logger.error(f"Reopening editor failed: {e}")
            self.base.notifications.error("Reopening editor failed: see logs")

    def open_editor(
        self, path: str = None, exists=True, load_file=True
    ) -> Editor | BaseEditor:
        """Open a new editor with the given path.

        Args:
            path (str): The path of the file to open.
            exists (bool, optional): Whether the file exists. Defaults to True.

        Returns:
            Editor: The opened editor."""

        if path:
            if self.is_open(path):
                return self.editorsbar.switch_tabs(path)
            if path in self.closed_editors:
                return self.add_editor(self.closed_editors[path])

        return self.add_editor(Editor(self, path, exists, load_file=load_file))

    def open_diff_editor(self, path: str, exists: bool) -> None:
        """Open a new diff editor with the given path.

        Args:
            path (str): The path of the file to open.
            exists (bool): Whether the file exists."""

        self.add_editor(Editor(self, path, exists, diff=True))

    def diff_files(self, file1: str, file2: str, standalone: bool = False) -> None:
        """Diff two files.

        Args:
            file1 (str): The path of the first file.
            file2 (str): The path of the second file."""

        self.add_editor(
            Editor(self, file1, True, file2, diff=True, standalone=standalone)
        )

    def open_game(self, id: str) -> None:
        """Open a new game editor with the given id.

        Args:
            id (str): The id of the game."""

        self.add_editor(Game(self, id))

    def close_editor(self, editor: Editor) -> None:
        """Closes the given editor.
        Keeps the editor in cache.

        Args:
            editor (Editor): The editor to close."""

        if editor in self.active_editors:
            self.active_editors.remove(editor)
        editor.grid_forget()
        self.refresh()

        if editor.content and editor.content.editable:
            self.base.language_server_manager.tab_closed(editor.content.text)

        self.closed_editors.append(editor)
        if len(self.closed_editors) > self.max_closed_editors:
            oldest_editor = self.closed_editors.pop(0)
            oldest_editor.destroy()

        self.base.open_editors.remove_item(editor)

    def restore_last_closed_editor(self) -> None:
        if self.closed_editors:
            editor = self.closed_editors.pop()
            self.add_editor(editor)
            # self.base.notifications.info(f"Restored {editor.filename}")
        else:
            self.base.notifications.info("No recently closed editors to restore")

    def close_editor_by_path(self, path: str) -> None:
        """Closes the editor with the given path.
        Keeps the editor in cache.

        Args:
            path (str): The path of the editor to close."""

        e = self.get_editor(path)
        self.close_editor(e)
        return e

    def get_editor(self, path: str) -> Editor:
        """Get editor by path.

        Args:
            path (str): The path of the editor to get."""

        for editor in self.active_editors:
            if editor.path == path:
                return editor

    def close_active_editor(self) -> None:
        self.editorsbar.close_active_tab()

    def delete_editor(self, editor: Editor) -> None:
        """Delete the given editor.

        Args:
            editor (Editor): The editor to delete."""

        if editor not in self.active_editors:
            return

        self.active_editors.remove(editor)
        self.editorsbar.delete_tab(editor)
        if editor.path in self.closed_editors:
            self.closed_editors.pop(editor.path)

        editor.destroy()
        self.base.open_editors.remove_item(editor)
        self.refresh()

    def set_active_editor(self, editor: Editor) -> Editor:
        """Set an existing editor to currently shown one.

        Args:
            editor (Editor): The editor to set as active."""

        for tab in self.editorsbar.active_tabs:
            if tab.editor == editor:
                self.editorsbar.set_active_tab(tab)
        self.base.open_editors.set_active(editor)
        self.refresh()
        return editor

    def set_active_editor_by_index(self, index: int) -> Editor:
        """Set an existing editor to currently shown one by index.

        Args:
            index (int): The index of the editor to set as active."""

        if index < len(self.editorsbar.active_tabs):
            self.editorsbar.set_active_tab(self.editorsbar.active_tabs[index])
            return self.editorsbar.active_tabs[index].editor

    def set_active_editor_by_path(self, path: str) -> Editor:
        """Set an existing editor to currently shown one by path.

        Args:
            path (str): The path of the editor to set as active."""

        for tab in self.editorsbar.active_tabs:
            if (
                tab.editor.path
                and path
                and (os.path.abspath(tab.editor.path) == os.path.abspath(path))
            ):
                self.editorsbar.set_active_tab(tab)
                return tab.editor

    def split_editor(self) -> None:
        # TODO: Implement split editor
        ...

    def change_tab_forward(self) -> None:
        self.editorsbar.change_tab_forward()

    def change_tab_back(self) -> None:
        self.editorsbar.change_tab_back()

    @property
    def active_editor(self) -> Editor:
        if not self.editorsbar.active_tab:
            return

        return self.editorsbar.active_tab.editor

    def refresh(self) -> None:
        if not self.active_editors:
            self.editorsbar.clear_buttons()
            self.emptytab.grid()
            self.editorsbar.active_tabs.clear()
            self.base.set_title(
                os.path.basename(self.base.active_directory)
                if self.base.active_directory
                else None
            )
            self.editorsbar.hide_tab_container()
        else:
            self.editorsbar.show_tab_container()
            self.emptytab.grid_remove()

        self.base.update_statusbar()



import tkinter as tk

from biscuit.common import Menu


class SourceControlMenu(Menu):
    def get_coords(self, e: tk.Event):
        return e.widget.winfo_rootx(), e.widget.winfo_rooty() + e.widget.winfo_height()



from __future__ import annotations

import tkinter as tk
import typing

if typing.TYPE_CHECKING:
    from .root import Root


class Grip(tk.Frame):
    """A grip to resize the application

    - Can be placed on the left, right, top, or bottom of the application
    """

    def __init__(self, base: Root, side: str, cursor: str, *args, **kwargs) -> None:
        super().__init__(base, *args, **kwargs)
        self.base = base
        self.side = side
        self.config(bg=self.base.base.theme.border, cursor=cursor)
        self.bind("<B1-Motion>", lambda _: self.base.base.resize(self.side))



import os
import platform
import subprocess
import tkinter as tk

from biscuit.common.actionset import ActionSet
from biscuit.common.icons import Icons

from ..panelview import PanelView
from .menu import TerminalMenu
from .shells import SHELLS, Default
from .tabs import Tabs
from .terminalbase import TerminalBase


def get_home_directory() -> str:
    if os.name == "nt":
        return os.path.expandvars("%USERPROFILE%")
    if os.name == "posix":
        return os.path.expanduser("~")
    return "."


class Terminal(PanelView):
    """Manages all integrated terminal instances.

    Tabbable terminal instances. Can spawn multiple terminals."""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)

        self.config(bg=self.base.theme.border)
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)
        self.grid_propagate(False)

        self.addmenu = TerminalMenu(self, "addterminal")
        for i in SHELLS:
            self.addmenu.add_command(i.name, lambda i=i: self.open_shell(i))

        self.menu = TerminalMenu(self, "terminal")
        self.menu.add_command("Clear Terminal", self.clear_terminal)

        self.__actions__ = [
            (Icons.ADD, self.addmenu.show),
            (Icons.TRASH, self.delete_active_terminal),
            (Icons.ELLIPSIS, self.menu.show),
        ]

        self.tabs = Tabs(self)
        self.tabs.grid(row=0, column=1, padx=(1, 0), sticky=tk.NS)

        self.active_terminals = []

        self.run_actionset = ActionSet(
            "Run Command in Terminal",
            "runc:",
            pinned=[["Run command? {}", self.run_command]],
        )
        self.base.palette.register_actionset(lambda: self.run_actionset)

    def add_default_terminal(self) -> Default:
        default_terminal = Default(
            self, cwd=self.base.active_directory or get_home_directory()
        )
        self.add_terminal(default_terminal)
        return default_terminal

    def add_current_terminal(self, *_) -> None:
        "Spawns an instance of currently active terminal"
        self.add_terminal(
            self.active_terminal_type(
                self, cwd=self.base.active_directory or get_home_directory()
            )
        )

    def add_terminals(self, terminals) -> None:
        "Append terminals to list. Create tabs for them."
        for terminal in terminals:
            self.add_terminal(terminal)

    def add_terminal(self, terminal) -> None:
        "Appends a terminal to list. Create a tab."
        self.active_terminals.append(terminal)
        self.tabs.add_tab(terminal)

    def open_shell(self, shell) -> None:
        self.add_terminal(
            shell(self, cwd=self.base.active_directory or get_home_directory())
        )

    def open_terminal(self, path=None) -> None:
        """Open another instance of active terminal in the current directory.
        If no active terminal, open a default terminal."""

        self.add_terminal(
            self.active_terminal_type(
                self, cwd=path or self.base.active_directory or get_home_directory()
            )
        )

    def delete_all_terminals(self) -> None:
        "Permanently delete all terminals."
        for terminal in self.active_terminals:
            terminal.destroy()

        self.tabs.clear_all_tabs()
        self.active_terminals.clear()

    def delete_terminal(self, terminal) -> None:
        "Permanently delete a terminal."
        terminal.destroy()
        self.active_terminals.remove(terminal)

    def delete_active_terminal(self) -> None:
        "Closes the active tab"
        try:
            self.tabs.close_active_tab()
        except IndexError:
            pass

    def set_active_terminal(self, terminal) -> None:
        "set an existing terminal to currently shown one"
        for tab in self.tabs.tabs:
            if tab.terminal == terminal:
                self.tabs.set_active_tab(tab)

    def clear_terminal(self, *_) -> None:
        if active := self.active_terminal:
            active.clear()

    def run_command(self, command: str) -> None:
        if not self.active_terminal:
            default = self.add_default_terminal()
            default.run_command(command)
            # this won't work, TODO: implement a queue for commands
        else:
            self.active_terminal.run_command(command)

    def run_external_console(self, command: str) -> None:
        "Run a command in external console."
        match platform.system():
            case "Windows":
                subprocess.Popen(["start", "cmd", "/K", command], shell=True)
            case "Linux":
                subprocess.Popen(["x-terminal-emulator", "-e", command])
            case "Darwin":
                subprocess.Popen(["open", "-a", "Terminal", command])
            case _:
                self.base.notifications.warning("No terminal emulator detected.")

    # TODO: Implement these
    def open_pwsh(self): ...

    def open_cmd(self): ...

    def open_python(self): ...

    @property
    def active_terminal_type(self):
        if active := self.active_terminal:
            return type(active)

        return Default

    @property
    def active_terminal(self) -> TerminalBase:
        "Get active terminal."
        if not self.tabs.active_tab:
            return

        return self.tabs.active_tab.terminal

    def refresh(self) -> None:
        if not self.active_terminals:
            self.master.toggle_panel()



import os
import tkinter as tk

from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, IconLabelButton, Label, LinkLabel, Shortcut
from biscuit.editor.misc.quickitem import QuickItem, RecentItem

from ..editorbase import BaseEditor


class Welcome(BaseEditor):
    name = "welcome"

    def __init__(self, master, exists=False, editable=False, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(padx=100, pady=50, **self.base.theme.editors)

        self.filename = "Welcome"

        self.container = Frame(self, **self.base.theme.editors)
        self.container.place(relx=0.5, rely=0.5, anchor=tk.CENTER)

        # try:
        #     self.logo = Label(
        #         self.container,
        #         image=self.base.settings.resources.logo.subsample(100, 100),
        #         **self.base.theme.editors.biscuit_labels
        #     )
        #     self.logo.grid(row=0, column=0, sticky=tk.NSEW)
        # except tk.TclError:
        #     pass

        self.title = Label(
            self.container,
            text="BISCUIT",
            font=("Fira Code", int(60 * self.base.scale), "bold"),
            fg=self.base.theme.biscuit,
            **self.base.theme.editors.biscuit_labels
        )
        self.title.pack(fill=tk.BOTH, padx=20)

        self.create_quick_group()

        # TODO: make this optional
        # self.create_recent_group()

    def create_quick_group(self):
        quick = Frame(self.container, **self.base.theme.editors)
        quick.pack(fill=tk.BOTH, expand=True, pady=(20, 0))

        QuickItem(
            quick,
            "New File",
            Icons.NEW_FILE,
            self.base.commands.new_file,
            ["Ctrl", "N"],
        ).pack(fill=tk.X, expand=True)

        QuickItem(
            quick,
            "Open File...",
            Icons.FOLDER_OPENED,
            self.base.commands.open_file,
            ["Ctrl", "O"],
        ).pack(fill=tk.X, expand=True)

        QuickItem(
            quick,
            "Open Folder...",
            Icons.FOLDER,
            self.base.commands.open_directory,
            ["Ctrl", "Shift", "O"],
        ).pack(fill=tk.X, expand=True)

        # TODO add following shortcuts
        QuickItem(
            quick,
            "Restore recent session",
            Icons.HISTORY,
            self.base.commands.restore_recent_session,
            ["Ctrl", "Alt", "R"],
        ).pack(fill=tk.X, expand=True)

        QuickItem(
            quick,
            "Config",
            Icons.SETTINGS,
            self.base.commands.open_settings,
            ["Ctrl", ","],
        ).pack(fill=tk.X, expand=True)

        QuickItem(
            quick,
            "Extensions",
            Icons.EXTENSIONS,
            self.base.commands.show_extensions,
            ["Ctrl", "Shift", "X"],
        ).pack(fill=tk.X, expand=True)

    def create_recent_group(self):
        Label(
            self.container,
            text="Recently opened",
            font=("Fira Code", 15),
            **self.base.theme.editors.labels
        ).pack(pady=(40, 0), anchor=tk.W)
        recents = Frame(self.container, **self.base.theme.editors)
        recents.pack(fill=tk.BOTH, expand=True, padx=5)
        recents.grid_columnconfigure(0, weight=1)

        for i, p in enumerate(self.base.history.folder_history.list):
            RecentItem(recents, os.path.basename(p[0]), p[1]).grid(
                row=i, column=0, sticky=tk.EW, pady=2
            )



from __future__ import annotations

import tkinter as tk
import typing
from tkinter.messagebox import askyesno

from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, IconButton
from biscuit.editor import BreadCrumbs

from .menu import EditorsbarMenu
from .tab import Tab
from .tabscroll import TabScroll

if typing.TYPE_CHECKING:
    from biscuit.editor import Editor

    from .manager import EditorsManager


class EditorsBar(Frame):
    """Editors Bar for Editors

    - Manages the tabs
    - Manages the action buttons for the tabs
    """

    def __init__(self, master: EditorsManager, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.layout.content.editors.bar)
        self.master: EditorsManager = master

        self.major_container = Frame(self, **self.base.theme.layout.content.editors.bar)
        self.major_container.pack(fill=tk.BOTH, expand=True)

        self.active_tabs: list[Tab] = []
        self.active_tab = None

        self.tab_control_container = TabScroll(self.major_container)
        self.tab_control_container.pack(fill=tk.Y, side=tk.LEFT)

        self.tab_container = Frame(self.major_container, bg=self.base.theme.border)
        self.tab_container.pack(fill=tk.BOTH, side=tk.LEFT)

        self.menu = EditorsbarMenu(self.major_container, "tabs")
        self.menu.add_command(
            "Show Opened Editors", lambda: self.base.palette.show("active:")
        )
        self.menu.add_command(
            "Restore Last Closed Editor", self.base.commands.restore_last_closed_editor
        )
        self.menu.add_separator(10)
        self.menu.add_command("Close All", self.master.delete_all_editors)

        self.buttons: list[IconButton] = []
        self.default_buttons = (
            (Icons.ELLIPSIS, self.menu.show),
            (Icons.CHEVRON_DOWN, self.base.open_editors.show),
            (Icons.ADD, self.base.commands.open_empty_editor),
        )

        self.action_container = Frame(
            self.major_container, **self.base.theme.layout.content.editors.bar
        )
        self.action_container.pack(fill=tk.BOTH, side=tk.RIGHT, padx=(0, 10))

        for button in self.default_buttons:
            if isinstance(button, list | tuple):
                IconButton(
                    self.action_container, iconsize=12, pady=6, hfg_only=True, *button
                ).pack(side=tk.RIGHT, fill=tk.Y)
            else:
                IconButton(self.action_container, **button).pack(
                    side=tk.RIGHT, fill=tk.Y
                )

        self.secondary_container = Frame(
            self, **self.base.theme.layout.content.editors.bar
        )
        self.secondary_container.pack(fill=tk.BOTH, expand=True)

        self.breadcrumbs = BreadCrumbs(self.secondary_container)
        self.breadcrumbs.show()

    def change_tab_forward(self) -> None:
        if self.active_tab:
            i = self.active_tabs.index(self.active_tab)
            next_index = (i + 1) % len(self.active_tabs)
            self.active_tabs[next_index].select()

    def change_tab_back(self) -> None:
        if self.active_tab:
            i = self.active_tabs.index(self.active_tab)
            prev_index = (i - 1) % len(self.active_tabs)
            self.active_tabs[prev_index].select()

    def hide_breadcrumbs(self) -> None:
        self.secondary_container.pack_forget()

    def show_breadcrumbs(self) -> None:
        self.breadcrumbs.clear()
        self.secondary_container.pack(fill=tk.BOTH, expand=True)

    def hide_tab_container(self) -> None:
        self.tab_container.pack_forget()

    def show_tab_container(self) -> None:
        self.tab_container.pack(
            fill=tk.BOTH, side=tk.LEFT, before=self.action_container
        )

    def add_buttons(self, buttons: list[IconButton]) -> None:
        for button in buttons:
            button.pack(side=tk.LEFT)
            self.buttons.append(button)

    def replace_buttons(self, buttons: list[IconButton]) -> None:
        self.clear_buttons()
        self.add_buttons(buttons)

    def clear_buttons(self) -> None:
        for button in self.buttons:
            button.pack_forget()
        self.buttons.clear()

    def add_tab(self, editor: Editor) -> None:
        tab = Tab(self, editor)
        tab.pack(fill=tk.Y, side=tk.LEFT, padx=(0, 1), in_=self.tab_container)
        self.active_tabs.append(tab)

        tab.select()

    def close_active_tab(self) -> None:
        if self.active_tab:
            self.close_tab(self.active_tab)

    def save_unsaved_changes(self, e: Editor) -> None:
        if e.content and e.content.editable and e.content.unsaved_changes:
            if askyesno(
                f"Unsaved changes",
                f"Do you want to save the changes you made to {e.filename}",
            ):
                if e.exists:
                    e.save()
                else:
                    self.base.commands.save_file_as()
                print(f"Saved changes to {e.path}.")

    def close_tab(self, tab: Tab) -> None:
        if e := tab.editor:
            self.save_unsaved_changes(e)

        try:
            i = self.active_tabs.index(tab)
        except ValueError:
            # most probably in case of diff editors, not handled
            return

        was_selected = tab == self.active_tab

        assert self.active_tabs.pop(i) == tab
        self.master.close_editor(tab.editor)
        tab.destroy()

        if not was_selected:
            return

        self.hide_breadcrumbs()
        if self.active_tabs:
            if i < len(self.active_tabs):
                self.active_tabs[i].select()
            else:
                self.active_tabs[i - 1].select()
        else:
            self.active_tab = None

    def close_tab_helper(self, editor: Editor) -> None:
        for tab in self.active_tabs:
            if tab.editor == editor:
                try:
                    self.active_tabs.remove(tab)
                    tab.destroy()
                except ValueError:
                    return

    def delete_tab(self, editor: Editor):
        self.hide_breadcrumbs()

        for tab in self.active_tabs:
            if tab.editor == editor:
                self.active_tabs.remove(tab)
                tab.destroy()
                return

    def set_active_tab(self, selected_tab: Tab) -> None:
        self.active_tab = selected_tab
        if selected_tab.editor.content:
            self.replace_buttons(selected_tab.editor.content.__buttons__)
        else:
            self.clear_buttons()
        for tab in self.active_tabs:
            if tab != selected_tab:
                tab.deselect()
        self.master.refresh()

    def clear_all_tabs(self) -> None:
        for tab in self.active_tabs:
            tab.destroy()

        self.active_tabs.clear()
        self.active_tab = None

    def switch_tabs(self, path: str) -> None:
        for tab in self.active_tabs:
            if tab.editor.path == path:
                tab.select()
                return tab.editor

    def scroll_left(self) -> None: ...
    def scroll_right(self) -> None: ...



class ChatModelInterface:
    """Interface for chat models.

    Must implement the following methods:
        - start_chat
        - send_message
    """

    def __init__(self, api_key: str = "", prompt=""):
        self.api_key = api_key
        self.model = None
        self.chat = None
        self.prompt = prompt

    def set_api_key(self, api_key: str):
        self.api_key = api_key

    def set_prompt(self, prompt: str):
        self.prompt = prompt

    def start_chat(self): ...
    def send_message(self, message: str): ...



from __future__ import annotations

import os
import tkinter as tk
import typing
from collections import defaultdict

from biscuit.common.ui import Frame, Label, Toplevel

from ..text import TextEditor
from .tree import PeekTree

if typing.TYPE_CHECKING:
    from biscuit.editor.text import Text
    from biscuit.language.data import Jump, JumpLocationRange


class Peek(Toplevel):
    """Floating window for peeking definitions, declarations, and references.

    Peek is triggered only by the LSP client, and it shows a list of locations
    where the symbol is defined, declared, or referenced.
    There is a preview pane for peeking the content of the selected location
    without switching tabs. Hence it's called peek!

    Peek fills the width of the tab and is placed right below the cursor.
    TODO: Peek renders on top of the tab, so the lines behind it are not visible.
    So skip the space taken by peek's height and render lines below it.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(padx=0, pady=1, bg=self.base.theme.biscuit)
        self.withdraw()
        self.overrideredirect(True)
        self.wm_attributes("-topmost", True)

        container = Frame(self, **self.base.theme.editors)
        container.pack(fill=tk.X, pady=(0, 1))

        self.filename = Label(
            container,
            font=self.base.settings.uifont,
            **self.base.theme.editors.labels,
            anchor=tk.W,
        )
        self.filename.pack(fill=tk.X, side=tk.LEFT)
        self.path = Label(
            container,
            font=self.base.settings.uifont,
            **self.base.theme.editors.labels,
            anchor=tk.W,
        )
        self.path.config(fg=self.base.theme.primary_foreground)
        self.path.pack(fill=tk.X, side=tk.LEFT)

        self.tree = PeekTree(
            self, width=100, singleclick=self.switch_editor, doubleclick=self.choose
        )

        self.editor = TextEditor(self, standalone=True, minimalist=True)
        self.tree.pack(side=tk.LEFT, fill=tk.Y)
        self.editor.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

        self.latest_tab = None
        self.mousein = False
        self.active = False

        self.editor.text.bind("<Double-Button-1>", self.choose)
        self.bind("<Enter>", lambda _: setattr(self, "mousein", True))
        self.bind("<Leave>", lambda _: setattr(self, "mousein", False))
        self.bind("<FocusOut>", self.hide)
        self.bind("<Escape>", self.hide)

    def refresh_geometry(self, pos: str, tab: Text):
        """Refresh the geometry of the window.

        The window is placed right below the cursor, and the widget fills the
        width of the tab.

        Args:
            pos (str): The position of the cursor.
            tab (Text): The current tab."""

        tab.update_idletasks()

        pos_x, pos_y = tab.master.winfo_rootx(), tab.master.winfo_rooty()

        pos = tab.index(pos + " linestart")
        bbox = tab.bbox(pos)
        if not bbox:
            return

        bbx_x, bbx_y, _, bbx_h = bbox
        self.geometry(
            "{}x{}+{}+{}".format(
                tab.master.winfo_width(), 250, pos_x, pos_y + bbx_y + bbx_h
            )
        )

    def show(self, tab: Text, jump: Jump):
        """Show the definitions window.

        Args:
            tab (Text): The current tab.
            jump (Jump): The Jump object."""

        self.update_locations(jump.locations)
        if jump.locations:
            path = jump.locations[0].file_path
            self.editor.text.load_new_file(path)
            self.editor.bind(
                "<<FileLoaded>>",
                lambda _, pos=jump.locations[0].start: self.refresh_and_goto(pos),
            )
            self.filename.set_text(os.path.basename(path))
            self.path.set_text(os.path.dirname(path))

        self.active = True
        self.latest_tab = tab

        self.refresh_geometry(jump.pos, tab)
        self.deiconify()
        self.lift()
        self.focus_set()

    def hide(self, *_):
        """Hide the definitions window."""

        if self.mousein:
            return

        self.active = False
        self.withdraw()
        self.clear()
        self.editor.clear()

    def force_hide(self):
        """Force hide peek."""

        self.active = False
        self.withdraw()
        self.clear()
        self.editor.clear()

    def clear(self):
        """Clear the list of locations."""

        self.tree.delete(*self.tree.get_children())

    def switch_editor(self, *_):
        """Switch the editor to the selected item."""

        try:
            path, start = self.tree.item(self.tree.focus())["values"]
        except ValueError:
            return

        self.editor.text.load_new_file(path)
        self.editor.bind(
            "<<FileLoaded>>", lambda _, pos=start: self.refresh_and_goto(pos)
        )
        self.filename.set_text(os.path.basename(path))
        self.path.set_text(os.path.dirname(path))

    def refresh_and_goto(self, pos: str):
        """Refresh the editor and go to the position.

        Args:
            pos (str): The position to go to."""

        self.editor.text.focus_set()
        self.editor.text.goto(pos)
        self.editor.text.refresh()

    def choose(self, *_):
        """Choose the selected item and go to the location."""

        try:
            path, start = self.tree.item(self.tree.focus())["values"]
        except ValueError:
            return

        self.base.goto_location(path, start)
        self.force_hide()

    def update_locations(self, locations: list[JumpLocationRange]) -> None:
        """Update the list of locations."""

        groups: dict[str, list[JumpLocationRange]] = defaultdict(list)
        for location in locations:
            groups[location.file_path].append(location)

        for path, locations in groups.items():
            if len(locations) > 1:
                i = self.tree.insert(
                    "",
                    "end",
                    text=os.path.basename(path),
                    values=(path, locations[0].start),
                    open=True,
                    image="document",
                )
                for loc in locations:
                    self.tree.insert(i, "end", text=loc.start, values=(path, loc.start))
            else:
                self.tree.insert(
                    "",
                    "end",
                    text=os.path.basename(path),
                    values=(path, locations[0].start),
                    image="document",
                )



import tkinter as tk

from biscuit.common.ui import Text


class TerminalText(Text):
    """TerminalText is a Text widget that is used to display the terminal
    output and to get the user input.

    It is a proxy widget that limits the editable area to text after the input mark
    and prevents deletion before the input mark.

    Attributes:
        proxy_enabled (bool): flag to enable/disable the proxy.
        _history (list): list of commands entered by the user.
        _history_level (int): current index of the history list.
    """

    def __init__(self, master=None, proxy_enabled: bool = True, **kw) -> None:
        """Initializes the TerminalText widget.

        Args:
            master (tk.Widget): parent widget.
            proxy_enabled (bool): flag to enable/disable the proxy.
            **kw: arbitrary keyword arguments."""

        super().__init__(master, **kw)
        self.mark_set("input", "insert")
        self.mark_gravity("input", "left")

        self.proxy_enabled = proxy_enabled
        self.config(**self.base.theme.editors.text, highlightthickness=0)

        self._history = []
        self._history_level = 0
        self.bind("<Up>", self.history_up)
        self.bind("<Down>", self.history_down)

        self._orig = self._w + "_orig"
        self.tk.call("rename", self._w, self._orig)
        self.tk.createcommand(self._w, self._proxy)

    def history_up(self, *_) -> None:
        "moves up the history"

        if not self._history:
            return "break"

        self._history_level = max(self._history_level - 1, 0)
        self.mark_set("insert", "input")
        self.delete("input", "end")
        self.insert("input", self._history[self._history_level])

        return "break"

    def history_down(self, *_) -> None:
        "moves down the history"

        if not self._history:
            return "break"

        self._history_level = min(self._history_level + 1, len(self._history) - 1)
        self.mark_set("insert", "input")
        self.delete("input", "end")
        self.insert("input", self._history[self._history_level])

        return "break"

    def register_history(self, command: str) -> None:
        "registers a command in the history"

        # don't register empty commands or duplicates
        if command.strip() and (
            not self._history or command.strip() != self._history[-1]
        ):
            self._history.append(command.strip())
        self._history_level = len(self._history)

    def clear(self, *_) -> None:
        "clears all previously used commands and results"

        self.proxy_enabled = False

        lastline = self.get("input linestart", "input")
        self.delete("1.0", "end")
        self.insert("end", lastline)

        self.proxy_enabled = True

    def _proxy(self, *args) -> None:
        """proxy limits the editable area to text after the input mark
        and prevents deletion before the input mark.
        """

        if not self.proxy_enabled:
            return self.tk.call((self._orig,) + args)

        try:
            largs = list(args)
            if args[0] == "insert":
                if self.compare("insert", "<", "input"):
                    self.mark_set("insert", "end")
            elif args[0] == "delete":
                if self.compare(largs[1], "<", "input"):
                    if len(largs) == 2:
                        return
                    largs[1] = "input"

            result = self.tk.call((self._orig,) + tuple(largs))
            return result
        except:
            # most probably some tkinter-unhandled exception
            pass



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common import Icons
from biscuit.common.ui import Bubble, Menubutton

from .menu import ActionMenu

if typing.TYPE_CHECKING:
    from .activitybar import ActivityBar


class ActionMenuButton(Menubutton):
    """Menu buttons for the activity bar

    Menu buttons are used to open a menu in the activity bar.
    Action menu instances are attached to these buttons."""

    def __init__(
        self, master: ActivityBar, icon: Icons, text: str, *args, **kwargs
    ) -> None:
        super().__init__(master, *args, **kwargs)
        self.bubble = Bubble(self, text=text)
        self.config(
            text=icon,
            relief=tk.FLAT,
            font=("codicon", 12),
            cursor="hand2",
            padx=5,
            pady=1,
            **self.base.theme.layout.sidebar.actionbar.slot,
        )
        self.pack(fill=tk.X, side=tk.TOP)

        self.menu = ActionMenu(self, icon)
        self.bind("<Button-1>", self.menu.show)
        self.bind("<Enter>", self.hover)
        self.bind("<Leave>", self.bubble.hide)

    def hover(self, *_) -> None:
        self.master.switch_menu(self.menu)
        self.bubble.show()



from .endpoint import Endpoint


class Notifications(Endpoint):
    """Notifications endpoint

    This class is used to interact with the notifications API.
    """

    def __init__(self, *a) -> None:
        super().__init__(*a)
        self.notifications = self.base.notifications

        self.info = self.show = self.notifications.info
        self.warn = self.warning = self.notifications.warning
        self.error = self.notifications.error
        self.notify = self.notifications.notify



import random
import tkinter as tk

from .gamebase import BaseGame


# a simple physics engine
class PhysicsObject:
    def __init__(self, canvas, x, y, width, height, color="black"):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.color = color
        self.canvas = canvas

        self.vel_x = 0
        self.vel_y = 0
        self.accel_x = 0
        self.accel_y = 0

        self.is_jumping = False

        self.shape = self.canvas.create_rectangle(
            self.x,
            self.y,
            self.x + self.width,
            self.y + self.height,
            fill=self.color,
            outline="",
        )

    def update(self):
        self.vel_x += self.accel_x
        self.vel_y += self.accel_y

        self.x += self.vel_x
        self.y += self.vel_y

        self.canvas.coords(
            self.shape, self.x, self.y, self.x + self.width, self.y + self.height
        )

    def jump(self, *_):
        if not self.is_jumping:
            self.is_jumping = True
            self.vel_y = -10

    def check_collision(self, other):
        if (
            self.x < other.x + other.width
            and self.x + self.width > other.x
            and self.y < other.y + other.height
            and self.y + self.height > other.y
        ):
            return True
        return False


class Whoops(BaseGame):
    """404 not found page"""

    name = "404"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config(bg="#22223b")
        self.gravity = 0.5
        self.canvas = tk.Canvas(
            self,
            width=800,
            height=400,
            bg="#22223b",
            borderwidth=0,
            highlightthickness=0,
        )
        self.canvas.pack(pady=30)

        self.canvas.create_text(
            390, 100, text="404", fill="#4a4e69", font=("Fixedsys", 50)
        )
        self.car = PhysicsObject(self.canvas, 100, 100, 60, 30, "#f2e9e4")
        self.ground = PhysicsObject(self.canvas, 0, 350, 800, 50, "#4a4e69")
        self.obstacles = []

        self.min_obstacle_gap = 2
        self.obstacle_spawnpoint = 800

        self.bind_all("<space>", self.car.jump)
        self.focus_set()

        self.update_game()
        self.update_obstacles()

    def apply_gravity(self, car):
        if not car.is_jumping:
            car.accel_y = self.gravity

    def update_game(self):
        self.car.update()
        self.apply_gravity(self.car)

        if self.car.y + self.car.height >= self.ground.y:
            self.car.y = self.ground.y - self.car.height
            self.car.vel_y = 0
            self.car.is_jumping = False

        self.car.check_collision(self.ground)
        self.after(10, self.update_game)

    def spawn_obstacle(self):
        obstacle_width = random.randint(40, 80)
        obstacle_height = random.randint(80, 200)
        obstacle_x = self.obstacle_spawnpoint
        obstacle_y = self.ground.y - obstacle_height

        obstacle = PhysicsObject(
            self.canvas,
            obstacle_x,
            obstacle_y,
            obstacle_width,
            obstacle_height,
            random.choice(["#9a8c98", "#4a4e69"]),
        )
        self.obstacles.append(obstacle)
        self.canvas.tag_lower(obstacle.shape)

    def update_obstacles(self):
        for obstacle in self.obstacles:
            obstacle.vel_x = -3
            obstacle.update()

            if obstacle.x + obstacle.width < 0:
                self.canvas.delete(obstacle.shape)
                self.obstacles.remove(obstacle)

        # spawn a new obstacle if there's enough gap between the last one and the right side of the canvas
        if (
            not self.obstacles
            or self.obstacle_spawnpoint
            - (self.obstacles[-1].x + self.obstacles[-1].width)
            >= self.min_obstacle_gap
        ):
            self.spawn_obstacle()

        self.after(10, self.update_obstacles)



from __future__ import annotations

import re
import tkinter as tk
import tkinter.font as tkfont
import typing

import tkextrafont as extra

from ..common import ActionSet, extract_commands, formalize_command
from .bindings import Bindings
from .config import Config
from .resources import Resources
from .styles import Style

if typing.TYPE_CHECKING:
    from ..app import App


URL = re.compile(r"^(?:http)s?:")


class Formattable(str):
    """For formatting github urls in the palette
    - If the term is a url, it will be returned as is
    - Otherwise, it will be formatted as a github url"""

    def format(self, term) -> str:
        default = "https://github.com/"
        if not term or URL.match(term):
            return super().format(term)

        return super().format(f"{default}{term}")


# TODO: functional settings editor
# TODO: load/store config in sqlite3 db
class Settings:
    """Settings for the application

    - Contains the configuration, style, resources, and bindings
    - Registers commands to the action set
    """

    def __init__(self, base: App) -> None:
        self.base: App = base

        self.config = Config(self)
        self.resources = Resources(self)
        self.bindings = Bindings(self)

        self.setup_font()
        self.style = Style(self)

        self.commands = []
        self.setup_icon()

    def register_command(self, command: str, callback: typing.Callable) -> None:
        """Registers a new palette command.

        Args:
            command (str): The command to be displayed.
            callback (typing.Callable): The callback function to be called.
        """
        self.commands.append((command, callback))
        self.generate_actionset()

    def generate_actionset(self) -> None:
        self._actionset = ActionSet(
            "Show and run commands",
            ">",
            self.commands + self.base.game_manager.get_games(),
        )

    def setup_icon(self) -> None:
        try:
            self.base.call(
                "wm",
                "iconphoto",
                self.base._w,
                tk.PhotoImage(file=self.resources.get_res_path_platform("icon.png")),
            )
        except tk.TclError:
            pass

    def setup_font(self) -> None:
        try:
            self.iconfont = extra.Font(
                file=self.resources.get_font("codicon.ttf"), family="codicon"
            )
        except tk.TclError:
            pass

        self.font = tkfont.Font(family=self.config.font[0], size=self.config.font[1])
        self.font_bold = tkfont.Font(
            family=self.config.font[0], size=self.config.font[1], weight="bold"
        )
        self.autocomplete_font = tkfont.Font(
            family=self.config.font[0], size=self.config.font[1] - 1
        )
        self.uifont = tkfont.Font(
            family=self.config.uifont[0], size=self.config.uifont[1]
        )
        self.uifont_bold = tkfont.Font(
            family=self.config.uifont[0], size=self.config.uifont[1], weight="bold"
        )

    def late_setup(self) -> None:
        """Configurations that require full initialization of editor"""

        self.commands = [
            (
                formalize_command(name),
                lambda _, method=method: method(self.base.commands),
            )
            for name, method in extract_commands(self.base.commands)
        ]

        self.generate_actionset()
        self.base.palette.register_actionset(lambda: self.actionset)

        clone_actionset = ActionSet(
            "Clone git repository",
            "clone:",
            pinned=[[Formattable("clone {}"), self.base.clone_repo]],
        )
        self.base.palette.register_actionset(lambda: clone_actionset)

        self.symbols_actionset = ActionSet("Go to symbol in editor", "@", [])
        self.base.palette.register_actionset(lambda: self.symbols_actionset)

    @property
    def actionset(self):
        return self._actionset



from __future__ import annotations

import re
import tkinter as tk
import typing

from biscuit.common.icons import Icons
from biscuit.common.ui import ButtonsEntry, Frame, IconButton, Toplevel

if typing.TYPE_CHECKING:
    from biscuit.common.ui import Text

from .results import FindResults


class FindReplace(Toplevel):
    """Floating find and replace window"""

    def __init__(self, base, *args, **kwargs) -> None:
        super().__init__(base, *args, **kwargs)
        self.offset = 10
        self.active = False
        self.overrideredirect(True)
        self.config(padx=1, pady=1, bg=self.base.theme.border)
        self.withdraw()

        self.text = None
        self.matchstring = None
        self.replacestring = None
        self.matches = None
        self.term = tk.StringVar()

        self.container = Frame(self, padx=5, pady=5, **self.base.theme.findreplace)
        self.container.pack(fill=tk.BOTH)
        self.container.grid_columnconfigure(0, weight=1)

        # find
        self.findbox = ButtonsEntry(
            self.container,
            hint="Find",
            textvariable=self.term,
            buttons=((Icons.CASE_SENSITIVE,), (Icons.WHOLE_WORD,), (Icons.REGEX,)),
        )
        self.findbox.grid(row=0, column=0, pady=2)

        self.results_count = FindResults(self.container)
        self.results_count.grid(row=0, column=1, sticky=tk.NSEW, padx=5, pady=2)

        buttons = Frame(self.container, **self.base.theme.findreplace)
        buttons.grid(row=0, column=2, sticky=tk.NSEW, pady=2)
        IconButton(buttons, Icons.ARROW_UP, self.prev_match).pack(side=tk.LEFT)
        IconButton(buttons, Icons.ARROW_DOWN, self.next_match).pack(side=tk.LEFT)
        IconButton(buttons, Icons.LIST_SELECTION).pack(
            side=tk.LEFT
        )  # TODO add finding from selection
        IconButton(buttons, Icons.CLOSE, self.hide).pack(side=tk.LEFT)

        # replace
        self.replacebox = ButtonsEntry(
            self.container, hint="Replace", buttons=((Icons.PRESERVE_CASE),)
        )
        self.replacebox.grid(row=1, column=0, sticky=tk.NSEW, pady=2)

        buttons = Frame(self.container, **self.base.theme.findreplace)
        buttons.grid(row=1, column=1, sticky=tk.NSEW, padx=5, pady=2)
        IconButton(buttons, Icons.REPLACE, self.replace).pack(side=tk.LEFT)
        IconButton(buttons, Icons.REPLACE_ALL, self.replace_all).pack(side=tk.LEFT)

        self.term.trace_add("write", self.find)

        self.base.bind("<FocusIn>", lambda *_: self.lift, add=True)
        self.base.bind("<Configure>", self._follow_root, add=True)

    def _follow_root(self, *_):
        if not self.active:
            return

        try:

            self.update_idletasks()
            x = (
                self.text.winfo_rootx()
                + self.text.winfo_width()
                - self.winfo_width()
                - self.offset
            )
            y = self.text.winfo_rooty()
            self.geometry(f"+{x}+{y}")
        except tk.TclError:
            # root was destroyed
            pass

    def show(self, text: Text):
        self.text = text
        self.active = True
        self.update_idletasks()

        if self.text.tag_ranges(tk.SEL):
            selection = self.text.get(tk.SEL_FIRST, tk.SEL_LAST)
            self.findbox.delete("0", "end")
            self.findbox.insert("0", selection)
            self.text.mark_set("insert", tk.SEL_FIRST)
            self.get_find_input()

        self._follow_root()
        self.deiconify()
        self.lift()
        self.findbox.focus()

    def hide(self, *_):
        self.active = False
        self.text.tag_remove("found", "1.0", "end")
        self.text.tag_remove("foundcurrent", "1.0", "end")
        self.withdraw()

    @property
    def current(self):
        if not self.text.count("1.0", self.text.index(tk.INSERT), "chars"):
            return 0
        return self.text.count("1.0", self.text.index(tk.INSERT), "chars")[0]

    def highlight_matches(self):
        self.text.tag_remove("found", "1.0", "end")
        self.text.tag_remove("foundcurrent", "1.0", "end")

        for pos, match in self.matches.items():
            start = match.start()
            end = match.end()
            self.text.tag_add("found", f"1.0+{start}c", f"1.0+{end}c")

        if self.is_on_match():
            self.highlight_current()

    def highlight_current(self):
        self.text.tag_remove("foundcurrent", "1.0", "end")

        current = self.current
        match = self.matches[current]

        start = match.start()
        end = match.end()
        self.text.tag_add("foundcurrent", f"1.0+{start}c", f"1.0+{end}c")

    def get_find_input(self):
        if self.findbox.get() == "":
            self.text.tag_remove("found", "1.0", "end")
            self.text.tag_remove("foundcurrent", "1.0", "end")
            return

        current = self.current
        self.matches = {}
        self.matchstring = self.findbox.get()
        self.re_ = re.compile(self.matchstring)

        for match in self.re_.finditer(self.text.get_all_text()):
            self.matches[match.start()] = match

        self.highlight_matches()
        self.text.mark_set("insert", f"1.0 + {current}c")
        self.results_count.show(len(self.matches))

    def find(self, *_):
        """Find all matches and highlight them"""
        try:
            self.text = self.base.editorsmanager.active_editor.content.text
        except AttributeError:
            return

        self.get_find_input()
        self.lift()

    def next_match(self, *_):
        """Moves the editor focus to the next match"""
        if self.findbox.get() != self.matchstring:
            self.get_find_input()

        matchpos = [i for i in sorted(self.matches.keys()) if i > self.current]

        if len(matchpos) > 0:
            next_index = f"1.0 + {matchpos[0]}c"
            self.text.mark_set("insert", next_index)
            self.text.see(next_index)
            self.highlight_current()
        elif len(self.matches) > 0:
            self.text.mark_set("insert", "1.0")

            if self.is_on_match():
                self.highlight_current()
            else:
                self.next_match()

        self.lift()
        self.text.focus()

    def prev_match(self, *_):
        """Moves the editor focus to the previous match"""
        if self.findbox.get() != self.matchstring:
            self.get_find_input()
        matchpos = [i for i in sorted(self.matches.keys()) if i < self.current]
        if len(matchpos) > 0:
            next_index = f"1.0 + {matchpos[-1]}c"
            self.text.mark_set("insert", next_index)
            self.text.see(next_index)
            self.highlight_current()
        elif len(self.matches) > 0:
            self.text.mark_set("insert", "end")
            self.prev_match()
        self.lift()
        self.text.focus()

    def replace(self, *_):
        """replaces current (in focus) match, removing the match and writing the replace string"""
        self.replacestring = self.replacebox.get()
        if self.findbox.get() != self.matchstring:
            self.get_find_input()
        if self.is_on_match():
            match = self.matches[self.current]
            self.text.delete(f"1.0 + {match.start()}c", f"1.0 + {match.end()}c")
            self.text.insert(f"1.0 + {self.current}c", self.replacestring)
            self.get_find_input()
        self.lift()
        self.text.focus()

    def is_on_match(self):
        """tells if the editor is currently pointing to a match"""
        return self.current in self.matches

    def replace_all(self, *_):
        """replaces all occurences of the string for the replace string, it will even replace partial words."""
        self.get_find_input()
        nmatches = len(self.matches)
        current = self.current
        self.text.mark_set("insert", "1.0")
        self.replace()
        for i in range(nmatches):
            self.next_match()
            self.replace()
        self.text.mark_set("insert", f"1.0 + {current}c")



from __future__ import annotations

import tkinter as tk
import typing
import webbrowser

import mistune
from tkinterweb import HtmlFrame

from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, Label, LinkLabel, Scrollbar
from biscuit.common.ui.buttons import Button, IconLabelButton
from biscuit.common.ui.native import Toplevel

if typing.TYPE_CHECKING:
    from biscuit.views.extensions.extension import ExtensionGUI


class ExtensionViewer(Toplevel):
    """Extension Viewer

    Extension Viewer displays the details of an extension.
    """

    def __init__(
        self,
        master,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(master, *args, **kwargs)
        self.ext: ExtensionGUI = None
        self.withdraw()
        self.config(bg=self.base.theme.border, padx=1, pady=1)
        self.overrideredirect(True)

        container = Frame(self, padx=10, pady=10)
        title = Frame(container)

        self.name = Label(title, text=f"Extension", fg=self.base.theme.biscuit)
        self.name.config(font=(self.base.settings.font["family"], 16, "bold"))
        self.version = Label(title, text=f"v1.0.0", fg=self.base.theme.border)
        self.version.config(font=(self.base.settings.font["family"], 10, "bold"))

        self.install = Button(
            title,
            "Install",
            padx=10,
        )
        self.install.config(font=("Segoi UI", 10), pady=2)

        title.pack(side=tk.TOP, fill=tk.X, pady=10)
        self.name.pack(side=tk.LEFT, fill=tk.X, padx=(0, 5))
        self.version.pack(side=tk.LEFT, fill=tk.X)
        self.install.pack(side=tk.RIGHT, fill=tk.X)

        Label(container, text="…" * 100, fg=self.base.theme.border).pack(
            side=tk.TOP, fill=tk.X, pady=5
        )

        self.author = Label(
            container, text=f"Tommy", fg=self.base.theme.primary_foreground
        )
        self.author.config(font=self.base.settings.uifont)
        self.description = Label(
            container, text=f"Hello, world!", fg=self.base.theme.primary_foreground
        )
        self.description.config(font=self.base.settings.uifont)

        self.author.pack(side=tk.TOP, anchor=tk.W, pady=5)
        self.description.pack(side=tk.TOP, anchor=tk.W, pady=5)

        container.pack(
            side=tk.LEFT, fill=tk.BOTH, expand=True, pady=(0, 5), padx=(0, 5)
        )

        self.bind("<Escape>", lambda _: self.hide())
        self.bind("<FocusOut>", lambda _: self.hide())

    def show(self, ext: ExtensionGUI) -> None:
        self.ext = ext
        self.name.config(text=ext.name)
        self.version.config(text=f"v0.1.0")
        self.author.config(text=f"by @{ext.author}")
        self.description.config(text=ext.description)

        if ext.installed:
            self.install.config(text="Installed", bg=self.base.theme.biscuit_dark)
            self.install.set_command(ext.remove_extension)
        else:
            self.install.config(text="Install", bg=self.base.theme.biscuit)
            self.install.set_command(ext.run_fetch_extension)

        self.deiconify()
        self.focus_set()

        self.geometry(
            f"500x300+{self.winfo_screenwidth() // 2 - 250}+{self.winfo_screenheight() // 2 - 150}"
        )

    def hide(self) -> None:
        self.withdraw()
        self.ext = None



from __future__ import annotations

import platform
import tkinter as tk
import typing

from biscuit.common import Menu
from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, IconButton
from biscuit.common.ui.bubble import Bubble

from .item import MenubarItem
from .notification import Notifications
from .searchbar import SearchBar

if typing.TYPE_CHECKING:
    ...


class Menubar(Frame):
    """Menubar of the application

    - Contains the MenubarItems
    - Manages the MenubarItems
    """

    def __init__(self, master: Frame, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.menus: list[MenubarItem] = []
        self.events = self.base.commands

        # structure of the menubar

        # |---left container---|---searchbar---|---right container---|

        self.container = Frame(self, **self.base.theme.layout.menubar)
        self.container.pack(side=tk.LEFT, fill=tk.BOTH)

        self.searchbar = SearchBar(self)
        self.searchbar.place(relx=0.5, rely=0.5, anchor=tk.CENTER)

        self.container_right = Frame(self, **self.base.theme.layout.menubar)
        self.container_right.pack(side=tk.RIGHT, fill=tk.BOTH)

        self.notifications = Notifications(self.container_right)
        self.notifications.pack(side=tk.LEFT, fill=tk.BOTH)

        # make this custom titlebar for windows
        if platform.system() == "Windows":
            close = IconButton(
                self.container_right,
                icon=Icons.CHROME_CLOSE,
                iconsize=12,
                padx=15,
                pady=8,
                event=self.events.quit_biscuit,
            )
            close.config(activebackground="#e81123", activeforeground="white")
            close.pack(side=tk.RIGHT, fill=tk.Y, padx=0)

            IconButton(
                self.container_right,
                icon=Icons.CHROME_MAXIMIZE,
                iconsize=12,
                icon2=Icons.CHROME_RESTORE,
                padx=15,
                pady=8,
                event=self.events.maximize_biscuit,
            ).pack(side=tk.RIGHT, fill=tk.Y, padx=0)
            IconButton(
                self.container_right,
                icon=Icons.CHROME_MINIMIZE,
                iconsize=12,
                padx=15,
                pady=8,
                event=self.events.minimize_biscuit,
            ).pack(side=tk.RIGHT, fill=tk.Y, padx=0)
            self.config_bindings()

        self.update_idletasks()

        self.config(**self.base.theme.layout.menubar)
        self.add_menus()

    def update_notifications(self) -> None:
        """Updates the notifications icon and description on the status bar."""

        n = self.base.notifications.count
        self.notifications.set_icon(Icons.BELL_DOT if n else Icons.BELL)
        self.notifications.bubble.change_text(
            f"{n} notifications" if n else "No notifications"
        )

    def change_title(self, title: str) -> None:
        """Change the title of the searchbar

        Args:
            title (str): Title of the searchbar"""

        self.searchbar.label.change_text(title)

    def config_bindings(self) -> None:
        self.bind("<Map>", self.events.window_mapped)
        self.bind("<Button-1>", self.start_move)
        self.bind("<ButtonRelease-1>", self.stop_move)
        self.bind("<B1-Motion>", self.moving)

    def start_move(self, event) -> None:
        self.x = event.x
        self.y = event.y

    def stop_move(self, _) -> None:
        self.x = None
        self.y = None

    def moving(self, event: tk.Event) -> None:
        x = event.x_root - self.x
        y = event.y_root - self.y
        self.base.geometry(f"+{x}+{y}")

    def add_menu(self, text: str) -> Menu:
        """Add a new menu to the menubar

        Args:
            text (str): Text of the menu"""

        new_menu = MenubarItem(self, text)
        new_menu.pack(side=tk.LEFT, fill=tk.BOTH, in_=self.container)
        self.menus.append(new_menu)

        return new_menu.menu

    def add_menus(self) -> None:
        self.add_file_menu()
        self.add_edit_menu()
        self.add_view_menu()
        self.add_help_menu()

    def add_file_menu(self) -> None:
        events = self.events

        self.file_menu = self.add_menu("File")
        self.file_menu.add_command("New File", events.new_file)
        self.file_menu.add_command("New Window", events.new_window)
        self.file_menu.add_separator()
        self.file_menu.add_command("Open File", events.open_file)
        self.file_menu.add_command("Open Folder", events.open_directory)
        self.file_menu.add_command("Open Recent File...", events.show_recent_files)
        self.file_menu.add_command("Open Recent Folder...", events.show_recent_folders)
        self.file_menu.add_separator()
        self.file_menu.add_command("Open workspace...", events.open_workspace)
        self.file_menu.add_command(
            "Add Folder to Workspace...", events.add_folder_to_workspace
        )
        self.file_menu.add_command("Save Workspace As...", events.save_workspace_as)
        self.file_menu.add_command("Close Workspace", events.close_workspace)
        self.file_menu.add_separator()
        self.file_menu.add_command("Save", events.save_file)
        self.file_menu.add_command("Save As...", events.save_file_as)
        self.file_menu.add_command("Save All", events.save_all)
        self.file_menu.add_separator()
        self.file_menu.add_command("Preferences", events.open_settings)
        self.file_menu.add_separator()
        self.file_menu.add_command("Close Editor", events.close_editor)
        self.file_menu.add_command("Close Folder", events.close_folder)
        self.file_menu.add_command("Close Window", events.quit_biscuit)
        self.file_menu.add_separator()
        self.file_menu.add_command("Exit", events.quit_biscuit)

    def add_edit_menu(self) -> None:
        events = self.events

        self.edit_menu = self.add_menu("Edit")
        self.edit_menu.add_command("Undo", events.undo)
        self.edit_menu.add_command("Redo", events.redo)
        self.edit_menu.add_separator()
        self.edit_menu.add_command("Cut", events.cut)
        self.edit_menu.add_command("Copy", events.copy)
        self.edit_menu.add_command("Paste", events.paste)
        self.edit_menu.add_separator()
        self.edit_menu.add_command("Find", events.find_symbol)
        self.edit_menu.add_command("Replace", events.replace_symbol)
        self.edit_menu.add_command("Find in Files", events.search_files)
        self.edit_menu.add_command("Go to Line/Column...", events.goto_line_column)
        self.edit_menu.add_separator()
        self.edit_menu.add_command("Change Language Mode", events.change_language_mode)
        self.edit_menu.add_checkable("Word Wrap", events.toggle_wordwrap)
        self.edit_menu.add_checkable("Block Cursor", events.toggle_block_cursor)
        self.edit_menu.add_checkable(
            "Toggle Relative Line Numbering", events.toggle_relative_line_numbering
        )

    def add_view_menu(self) -> None:
        events = self.events

        self.view_menu = self.add_menu("View")
        # TODO: Add the rest of the view menu items
        self.view_menu.add_command("Command Palette...", events.show_command_palette)
        self.view_menu.add_command("Explorer", events.show_directory_tree)
        self.view_menu.add_command("Outline", events.show_outline)
        self.view_menu.add_command("Search", events.show_search)
        self.view_menu.add_command("Source Control", events.show_source_control)
        self.view_menu.add_command("Extensions", events.show_extensions)
        self.view_menu.add_separator()
        self.view_menu.add_command("Terminal", events.show_terminal)
        self.view_menu.add_command("Log", events.show_logs)

    def add_help_menu(self) -> None:
        events = self.events

        self.help_menu = self.add_menu("Help")
        self.help_menu.add_command("Welcome", events.show_welcome)
        self.help_menu.add_command("Show All Commands", events.show_command_palette)
        self.help_menu.add_command("Documentation", events.open_biscuit_documentation)
        self.help_menu.add_command(
            "Show Release Notes", events.open_biscuit_release_notes
        )
        self.help_menu.add_separator()
        self.help_menu.add_command("Report Bug", events.report_bug)
        self.help_menu.add_command("Request Feature", events.request_feature)
        self.help_menu.add_separator()
        self.help_menu.add_command("View License", events.view_biscuit_licenses)
        self.help_menu.add_command(
            "Code of Conduct", events.open_biscuit_code_of_conduct
        )
        self.help_menu.add_separator()
        self.help_menu.add_command("About", events.show_about)

    def close_all_menus(self, *_) -> None:
        for btn in self.menus:
            btn.menu.hide()

    def switch_menu(self, item: MenubarItem) -> None:
        active = False
        for i in self.menus:
            if i.menu.active:
                active = True
            if i.menu != item.menu:
                i.menu.hide()
                i.deselect()

        if active:
            item.select()
            item.menu.show()

    def pack(self):
        super().pack(fill=tk.BOTH)



from __future__ import annotations

import typing
from typing import List

if typing.TYPE_CHECKING:
    from .loader import WorkspaceLoader


class Workspace:
    def __init__(self, loader: WorkspaceLoader, path: str, dirs: List[str]):
        self.loader = loader
        self.path = path
        self.dirs = dirs

    def add_dir(self, dir: str):
        self.dirs.append(dir)
        self.loader.dump_modified(self.path, self.dirs)

    def export(self):
        return {"dirs": self.dirs}



from __future__ import annotations

import os
import tkinter as tk
import typing

from pygments import lex
from pygments.lexers import get_lexer_by_name, get_lexer_for_filename
from pygments.style import Style

if typing.TYPE_CHECKING:
    from biscuit import App

    from .text import Text


class BiscuitStyle(Style):
    name = "biscuit"

    def __init__(self, master: Highlighter) -> None:
        self.base = master.base
        self.styles = self.base.theme.syntax
        self.background_color = self.base.theme.editors.background


class Highlighter:
    """Syntax Highlighter

    This highlighter uses pygments to highlight the text content based on the lexer provided.
    The lexer can be provided explicitly or it can be detected from the file extension.
    If the file extension is not recognized, it will default to plain text.

    Supported languages and text formats: https://pygments.org/docs/lexers/
    """

    def __init__(self, text: Text, language: str = None, *args, **kwargs) -> None:
        """Highlighter based on pygments lexers

        If language is not given, it will try to detect the language from the file extension.
        If the file extension is not recognized, it will default to plain text.

        Args:
            text (Text): The text instance to be highlighted
            language (str, optional): Language to highlight. Defaults to None."""

        self.text: Text = text
        self.base: App = text.base
        self.language = language

        if language:
            try:
                self.lexer = get_lexer_by_name(language)
                self.text.language = self.lexer.name
                self.text.language_alias = self.lexer.aliases[0]
            except:
                self.lexer = None
                self.text.language = "Plain Text"
                self.text.language_alias = "text"
                self.base.notifications.info("Selected lexer is not available.")
        else:
            try:
                if os.path.basename(text.path).endswith("txt"):
                    raise Exception()

                self.lexer = get_lexer_for_filename(
                    os.path.basename(text.path), encoding=text.encoding
                )
                self.text.language = self.lexer.name
                self.text.language_alias = self.lexer.aliases[0]
            except:
                self.lexer = None
                self.text.language = "Plain Text"
                self.text.language_alias = "text"

        self.tag_colors = self.base.theme.syntax
        self.setup_highlight_tags()

    def detect_language(self) -> None:
        """Detect the language from the file extension and set the lexer
        Refreshes language attribute of the text instance."""

        try:
            if os.path.basename(self.text.path).endswith("txt"):
                raise Exception()

            self.lexer = get_lexer_for_filename(
                os.path.basename(self.text.path), encoding=self.text.encoding
            )
            self.text.language = self.lexer.name
            self.text.language_alias = self.lexer.aliases[0]
            self.highlight()
        except:
            self.lexer = None
            self.text.language = "Plain Text"
            self.text.language_alias = "text"

    def change_language(self, language: str) -> None:
        """Change the language of the highlighter
        If language is not given, it will try to detect the language from the file extension.
        If the file extension is not recognized, it will default to plain text.

        Args:
            language (str): Language to highlight. Defaults to None."""

        try:
            self.lexer = get_lexer_by_name(language)
        except:
            self.lexer = None
            self.text.language = "Plain Text"
            self.text.language_alias = "text"
            self.base.notifications.info("Selected lexer is not available.")
            return

        self.text.language = self.lexer.name
        self.text.language_alias = self.lexer.aliases[0]
        self.tag_colors = self.base.theme.syntax
        self.text.master.on_change()
        self.base.statusbar.on_open_file(self.text)

    def setup_highlight_tags(self) -> None:
        """Setup the tags for highlighting the text content"""

        for token, props in self.tag_colors.items():
            if isinstance(props, dict):
                if "font" in props and isinstance(props["font"], dict):
                    f = self.base.settings.font.copy()
                    f.config(**props["font"])
                    props["font"] = f

                self.text.tag_configure(str(token), **props)
            else:
                self.text.tag_configure(str(token), foreground=props)

    def clear(self) -> None:
        """Clears the highlighting of the text content"""

        for token, _ in self.tag_colors.items():
            self.text.tag_remove(str(token), "1.0", tk.END)

    def highlight(self) -> None:
        """Highlight the text content

        This method highlights the text content based on the lexer provided.

        TODO: As of now, it highlights the entire text content.
        It needs to be optimized to highlight only the visible area."""

        if not self.lexer or not self.tag_colors:
            return

        for token, _ in self.tag_colors.items():
            self.text.tag_remove(str(token), "1.0", tk.END)

        text = self.text.get_all_text()

        # NOTE:  Highlighting only visible area
        # total_lines = int(self.text.index('end-1c').split('.')[0])
        # start_line = int(self.text.yview()[0] * total_lines)
        # first_visible_index = f"{start_line}.0"
        # last_visible_index =f"{self.text.winfo_height()}.end"
        # for token, _ in self.tag_colors.items():
        #     self.text.tag_remove(str(token), first_visible_index, last_visible_index)
        # text = self.text.get(first_visible_index, last_visible_index)

        self.text.mark_set("range_start", "1.0")
        for token, content in lex(text, self.lexer):
            self.text.mark_set("range_end", f"range_start + {len(content)}c")
            self.text.tag_add(str(token), "range_start", "range_end")
            self.text.mark_set("range_start", "range_end")

            # DEBUG
            # print(f"{content} is recognized as a <{str(token)}>")
        # print("==================================")



from __future__ import annotations

import typing

from biscuit.common import Menu

if typing.TYPE_CHECKING:
    from .variables import Variables


class VariablesContextMenu(Menu):
    def __init__(self, master: Variables, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master

        self.add_command("Copy Value", self.master.copy_value)
        self.add_command("Copy Expression", self.master.copy_expression)
        self.add_command("Copy Name", self.master.copy_name)

        self.add_separator()
        self.add_command("Set Value", self.master.set_value)

    def get_coords(self, e) -> list:
        return e.x_root, e.y_root

    def show(self, *e) -> None:
        super().show(*e)
